# coding=utf-8

import logging
import ctypes
import struct
import threading
import time
import functools
import random
import gc

### 调试日志

fmt_str = "%(asctime)s.%(msecs)03d - %(name)s - %(levelname)s - %(message)s"
log_formatter = logging.Formatter(fmt=fmt_str, datefmt='%H:%M:%S')

console_handler = logging.StreamHandler()
console_handler.setFormatter(log_formatter)
console_handler.setLevel(logging.INFO)

pvz_logger = logging.getLogger("pvz")
pvz_logger.addHandler(console_handler)
pvz_logger.setLevel(logging.INFO)


def enable_logging(on=True):
    """
    启用日志.

    输出调试信息开销较大, 会影响操作精度, 建议在调试完成后正式运行的时候关闭.

    @参数 on(bool): 是否启用, 默认启用.
    """
    pvz_logger.disabled = not on


def set_logging_level(level="INFO"):
    """
    设置日志级别.

    @参数 level(str): 可选值 ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"].
    """
    logging_level = {
        "DEBUG": logging.DEBUG,
        "INFO": logging.INFO,
        "WARNING": logging.WARNING,
        "ERROR": logging.ERROR,
        "CRITICAL": logging.CRITICAL,
    }
    pvz_logger.setLevel(logging_level[level])


# 频繁出现的调试信息
def debug(txt):
    pvz_logger.debug(txt)


# 常规调试信息
def info(txt):
    pvz_logger.info(txt)


# 不影响运行的警告
def warning(txt):
    pvz_logger.warning(txt)


# 用户操作出错
def error(txt):
    pvz_logger.error(txt)
    raise Exception(txt)


# 内部严重错误
def critical(txt):
    pvz_logger.critical(txt)
    raise Exception(txt)


### win32 动态库

user32 = ctypes.windll.user32
kernel32 = ctypes.windll.kernel32
winmm = ctypes.windll.winmm
gdi32 = ctypes.windll.gdi32

### win32 类型

from ctypes import c_size_t as SIZE_T
from ctypes.wintypes import BOOL
from ctypes.wintypes import DWORD
from ctypes.wintypes import INT
from ctypes.wintypes import UINT
from ctypes.wintypes import LONG
from ctypes.wintypes import HWND
from ctypes.wintypes import HANDLE
from ctypes.wintypes import HDC
from ctypes.wintypes import LPVOID
from ctypes.wintypes import LPCVOID
from ctypes.wintypes import LPCWSTR
from ctypes.wintypes import LPDWORD
from ctypes.wintypes import POINT
from ctypes.wintypes import LPPOINT
from ctypes.wintypes import RECT
from ctypes.wintypes import LPRECT
from ctypes.wintypes import WPARAM
from ctypes.wintypes import LPARAM


# typedef struct SECURITY_ATTRIBUTES {
#   DWORD  nLength;
#   LPVOID lpSecurityDescriptor;
#   BOOL   bInheritHandle;
# } SECURITY_ATTRIBUTES, *PSECURITY_ATTRIBUTES, *LPSECURITY_ATTRIBUTES;
class SECURITY_ATTRIBUTES(ctypes.Structure):
    _fields_ = [("nLength", DWORD), ("lpSecurityDescriptor", LPVOID), ("bInheritHandle", BOOL)]


LPSECURITY_ATTRIBUTES = ctypes.POINTER(SECURITY_ATTRIBUTES)

# typedef DWORD (__stdcall *LPTHREAD_START_ROUTINE) (
#     [in] LPVOID lpThreadParameter
# );
LPTHREAD_START_ROUTINE = ctypes.WINFUNCTYPE(DWORD, LPVOID)

### win32 函数接口 和 参数常量

# int MessageBoxW(
#   HWND    hWnd,
#   LPCWSTR lpText,
#   LPCWSTR lpCaption,
#   UINT    uType
# );
MessageBoxW = user32.MessageBoxW
MessageBoxW.argtypes = [HWND, LPCWSTR, LPCWSTR, UINT]
MessageBoxW.restype = INT

# HWND FindWindowW(
#   LPCWSTR lpClassName,
#   LPCWSTR lpWindowName
# );
FindWindowW = user32.FindWindowW
FindWindowW.argtypes = [LPCWSTR, LPCWSTR]
FindWindowW.restype = HWND

# DWORD GetWindowThreadProcessId(
#   HWND    hWnd,
#   LPDWORD lpdwProcessId
# );
GetWindowThreadProcessId = user32.GetWindowThreadProcessId
GetWindowThreadProcessId.argtypes = [HWND, LPDWORD]
GetWindowThreadProcessId.restype = DWORD

# HANDLE OpenProcess(
#   DWORD dwDesiredAccess,
#   BOOL  bInheritHandle,
#   DWORD dwProcessId
# );
OpenProcess = kernel32.OpenProcess
OpenProcess.argtypes = [DWORD, BOOL, DWORD]
OpenProcess.restype = HANDLE

PROCESS_ALL_ACCESS = 0x001F0FFF

# BOOL GetExitCodeProcess(
#   HANDLE  hProcess,
#   LPDWORD lpExitCode
# );
GetExitCodeProcess = kernel32.GetExitCodeProcess
GetExitCodeProcess.argtypes = [HANDLE, LPDWORD]
GetExitCodeProcess.restype = BOOL

STILL_ACTIVE = 0x00000103

# BOOL WINAPI CloseHandle(
#   _In_ HANDLE hObject
# );
CloseHandle = kernel32.CloseHandle
CloseHandle.argtypes = [HANDLE]
CloseHandle.restype = BOOL

# BOOL WINAPI ReadProcessMemory(
#   _In_  HANDLE  hProcess,
#   _In_  LPCVOID lpBaseAddress,
#   _Out_ LPVOID  lpBuffer,
#   _In_  SIZE_T  nSize,
#   _Out_ SIZE_T  *lpNumberOfBytesRead
# );
ReadProcessMemory = kernel32.ReadProcessMemory
ReadProcessMemory.argtypes = [HANDLE, LPCVOID, LPVOID, SIZE_T, LPDWORD]
ReadProcessMemory.restype = BOOL

# BOOL WINAPI WriteProcessMemory(
#   _In_  HANDLE  hProcess,
#   _In_  LPVOID  lpBaseAddress,
#   _In_  LPCVOID lpBuffer,
#   _In_  SIZE_T  nSize,
#   _Out_ SIZE_T  *lpNumberOfBytesWritten
# );
WriteProcessMemory = kernel32.WriteProcessMemory
WriteProcessMemory.argtypes = [HANDLE, LPVOID, LPCVOID, SIZE_T, LPDWORD]
WriteProcessMemory.restype = BOOL

# DWORD WINAPI GetLastError(void);
GetLastError = kernel32.GetLastError
GetLastError.argtypes = []
GetLastError.restype = DWORD

# HWND SetActiveWindow(
#   HWND hWnd
# );
SetActiveWindow = user32.SetActiveWindow
SetActiveWindow.argtypes = [HWND]
SetActiveWindow.restype = HWND

# BOOL SetForegroundWindow(
#   HWND hWnd
# );
SetForegroundWindow = user32.SetForegroundWindow
SetForegroundWindow.argtypes = [HWND]
SetForegroundWindow.restype = BOOL

# BOOL SetWindowPos(
#   HWND hWnd,
#   HWND hWndInsertAfter,
#   int  X,
#   int  Y,
#   int  cx,
#   int  cy,
#   UINT uFlags
# );
SetWindowPos = user32.SetWindowPos
SetWindowPos.argtypes = [HWND, HWND, INT, INT, INT, INT, UINT]
SetWindowPos.restype = BOOL

HWND_NOTOPMOST = -2
HWND_TOPMOST = -1
SWP_NOMOVE = 0x0002
SWP_NOSIZE = 0x0001
SWP_SHOWWINDOW = 0x0040

# LPVOID WINAPI VirtualAllocEx(
#   _In_     HANDLE hProcess,
#   _In_opt_ LPVOID lpAddress,
#   _In_     SIZE_T dwSize,
#   _In_     DWORD  flAllocationType,
#   _In_     DWORD  flProtect
# );
VirtualAllocEx = kernel32.VirtualAllocEx
VirtualAllocEx.argtypes = [HANDLE, LPVOID, SIZE_T, DWORD, DWORD]
VirtualAllocEx.restype = LPVOID

MEM_COMMIT = 0x00001000
PAGE_EXECUTE_READWRITE = 0x40

# BOOL WINAPI VirtualFreeEx(
#   _In_ HANDLE hProcess,
#   _In_ LPVOID lpAddress,
#   _In_ SIZE_T dwSize,
#   _In_ DWORD  dwFreeType
# );
VirtualFreeEx = kernel32.VirtualFreeEx
VirtualFreeEx.argtypes = [HANDLE, LPVOID, SIZE_T, DWORD]
VirtualFreeEx.restype = BOOL

MEM_RELEASE = 0x00008000

# HANDLE CreateRemoteThread(
#   HANDLE                 hProcess,
#   LPSECURITY_ATTRIBUTES  lpThreadAttributes,
#   SIZE_T                 dwStackSize,
#   LPTHREAD_START_ROUTINE lpStartAddress,
#   LPVOID                 lpParameter,
#   DWORD                  dwCreationFlags,
#   LPDWORD                lpThreadId
# );
CreateRemoteThread = kernel32.CreateRemoteThread
CreateRemoteThread.argtypes = [HANDLE, LPSECURITY_ATTRIBUTES, SIZE_T, LPTHREAD_START_ROUTINE, LPVOID, DWORD, LPDWORD]
CreateRemoteThread.restype = HANDLE

# DWORD WaitForSingleObject(
#   HANDLE hHandle,
#   DWORD  dwMilliseconds
# );
WaitForSingleObject = kernel32.WaitForSingleObject
WaitForSingleObject.argtypes = [HANDLE, DWORD]
WaitForSingleObject.restype = DWORD

INFINITE = 0xFFFFFFFF
WAIT_FAILED = 0xFFFFFFFF

# LRESULT SendMessageW(
#   HWND   hWnd,
#   UINT   Msg,
#   WPARAM wParam,
#   LPARAM lParam
# );
SendMessageW = user32.SendMessageW
SendMessageW.argtypes = [HWND, UINT, WPARAM, LPARAM]
SendMessageW.restype = LONG

# BOOL PostMessageW(
#   HWND   hWnd,
#   UINT   Msg,
#   WPARAM wParam,
#   LPARAM lParam
# );
PostMessageW = user32.PostMessageW
PostMessageW.argtypes = [HWND, UINT, WPARAM, LPARAM]
PostMessageW.restype = BOOL

WM_KEYDOWN = 0x0100
WM_KEYUP = 0x0101
VK_ESCAPE = 0x1B
VK_SPACE = 0x20
VK_RETURN = 0x0D
VK_LEFT = 0x25
VK_UP = 0x26
VK_RIGHT = 0x27
VK_DOWN = 0x28

WM_LBUTTONDOWN = 0x0201
WM_LBUTTONUP = 0x0202
WM_RBUTTONDOWN = 0x0204
WM_RBUTTONUP = 0x0205

MK_LBUTTON = 0x0001
MK_RBUTTON = 0x0002

# HDC GetDC(
#   HWND hWnd
# );
GetDC = user32.GetDC
GetDC.argtypes = [HWND]
GetDC.restype = HDC

# int GetDeviceCaps(
#   HDC hdc,
#   int index
# );
GetDeviceCaps = gdi32.GetDeviceCaps
GetDeviceCaps.argtypes = [HDC, INT]
GetDeviceCaps.restype = INT

HORZRES = 8
DESKTOPHORZRES = 118

# int ReleaseDC(
#   HWND hWnd,
#   HDC  hDC
# );
ReleaseDC = user32.ReleaseDC
ReleaseDC.argtypes = [HWND, HDC]
ReleaseDC.restype = INT

# MMRESULT timeBeginPeriod(
#   UINT uPeriod
# );
timeBeginPeriod = winmm.timeBeginPeriod
timeBeginPeriod.argtypes = [UINT]
timeBeginPeriod.restype = UINT

# MMRESULT timeEndPeriod(
#   UINT uPeriod
# );
timeEndPeriod = winmm.timeEndPeriod
timeEndPeriod.argtypes = [UINT]
timeEndPeriod.restype = UINT

# HANDLE GetCurrentProcess(
# );
GetCurrentProcess = kernel32.GetCurrentProcess
GetCurrentProcess.argtypes = []
GetCurrentProcess.restype = HANDLE

# DWORD GetPriorityClass(
#   HANDLE hProcess
# );
GetPriorityClass = kernel32.GetPriorityClass
GetPriorityClass.argtypes = [HANDLE]
GetPriorityClass.restype = DWORD

# BOOL SetPriorityClass(
#   HANDLE hProcess,
#   DWORD  dwPriorityClass
# );
SetPriorityClass = kernel32.SetPriorityClass
SetPriorityClass.argtypes = [HANDLE, DWORD]
SetPriorityClass.restype = BOOL

HIGH_PRIORITY_CLASS = 0x00000080
REALTIME_PRIORITY_CLASS = 0x00000100

### 进程操作

# 窗口句柄
pvz_hwnd = HWND()

# 进程标识
pvz_pid = DWORD()

# 进程句柄
pvz_handle = HANDLE()

# 游戏版本
pvz_version = None


def is_valid():
    """
    检查目标进程是否可用.

    @返回值 (bool): 未找到或者已经退出则返回 False.
    """

    if pvz_handle.value is None:
        return False

    exit_code = DWORD()
    GetExitCodeProcess(pvz_handle, ctypes.byref(exit_code))
    return exit_code.value == STILL_ACTIVE


def open_process_by_window(class_name, window_name):
    """
    根据窗口的类名和标题打开进程.

    @参数 class_name(str): 窗口类名, 可省略为 None.

    @参数 window_name(str): 窗口标题, 可省略为 None.

    @返回值 (bool): 成功打开目标进程则返回 True.
    """

    global pvz_hwnd, pvz_pid, pvz_handle

    # 关闭之前已经打开的句柄
    if is_valid():
        CloseHandle(pvz_handle)

    pvz_hwnd.value = None
    pvz_pid.value = 0
    pvz_handle.value = None

    pvz_hwnd.value = FindWindowW(class_name, window_name)
    if pvz_hwnd.value is not None:
        GetWindowThreadProcessId(pvz_hwnd, ctypes.byref(pvz_pid))
    if pvz_pid.value != 0:
        pvz_handle.value = OpenProcess(PROCESS_ALL_ACCESS, False, pvz_pid)

    result = pvz_handle.value is not None
    info("查找游戏窗口, 类名 '%s', 标题 '%s', 结果 %s." % (class_name, window_name, result))
    return result


def find_pvz():
    """
    查找原版植物大战僵尸游戏进程.

    @返回值 (bool): 查找成功返回 True, 没找到或是版本不符则返回 False.
    """

    global pvz_version

    # 不建议省略窗口类名, 因为可能存在其他标题相同的窗口从而引起查找失误.
    # 已知所有的植物大战僵尸一代电脑版的窗口类名均为 "MainWindow".
    # 原版英文版的窗口标题为 "Plants vs. Zombies".
    if not open_process_by_window("MainWindow", "Plants vs. Zombies"):
        open_process_by_window("MainWindow", None)

    if is_valid():
        if read_memory("unsigned int", 0x004140C5) == 0x0019B337:
            pvz_version = "1.0.0.1051"
            info("已找到游戏 1.0.0.1051 !!!")
            return True
        elif read_memory("unsigned int", 0x004140D5) == 0x0019B827:
            pvz_version = "1.2.0.1065"
            info("已找到游戏 1.2.0.1065 !!!")
            return True
        else:
            pvz_version = None
            warning("不支持的游戏版本 !!!")
            return False
    else:
        pvz_version = None
        warning("未找到游戏 !!!")
        return False


def pvz_ver():
    return pvz_version


# C/C++ 数据类型
cpp_typename = {
    "char": "b",
    "signed char": "b",
    "int8_t": "b",
    "unsigned char": "B",
    "byte": "B",
    "uint8_t": "B",
    "bool": "?",
    "short": "h",
    "int16_t": "h",
    "unsigned short": "H",
    "uint16_t": "H",
    "int": "i",
    "int32_t": "i",
    "intptr_t": "i",
    "unsigned int": "I",
    "uint32_t": "I",
    "uintptr_t": "I",
    "size_t": "I",
    "long": "l",
    "unsigned long": "L",
    "long long": "q",
    "int64_t": "q",
    "intmax_t": "q",
    "unsigned long long": "Q",
    "uint64_t": "Q",
    "uintmax_t": "Q",
    "float": "f",
    "double": "d",
}

# 读写内存时加锁
memory_lock = threading.Lock()

### 读写内存


def read_memory(data_type, *address, array=1):
    """
    读取内存数据.

    @参数 data_type(str): 数据类型, 取自 C/C++ 语言关键字, 可选值 ["char", "bool", "unsigned char", "short", "unsigned short", "int", "unsigned int", "long", "unsigned long", "long long", "unsigned long long", "float", "double"].

    @参数 address(int): 地址, 可为多级偏移.

    @参数 array(int): 数量. 默认一个, 大于一个时需要显式指定关键字参数.

    @返回值 (int/float/bool/tuple): 默认情况下返回单个数值, 获取多个数据则返回一个长度为指定数量的元组.

    @示例:

    >>> ReadMemory("int", 0x6a9ec0, 0x768, 0x5560)
    8000

    >>> ReadMemory("byte", 0x0041d7d0, array=3)
    (81, 131, 248)
    """

    if not is_valid():
        critical("目标进程不可用, 读内存失败.")

    memory_lock.acquire()

    level = len(address)  # 偏移级数
    offset = ctypes.c_void_p()  # 内存地址
    buffer = ctypes.c_uint()  # 中间数据缓冲
    bytes_read = ctypes.c_ulong()  # 已读字节数

    for i in range(level):
        offset.value = buffer.value + address[i]

        if i != level - 1:
            size = ctypes.sizeof(buffer)
            success = ReadProcessMemory(pvz_handle, offset, ctypes.byref(buffer), size, ctypes.byref(bytes_read))
            if success == 0 or bytes_read.value != size:
                critical("读取内存失败, 错误代码 %d." % GetLastError())

        else:
            fmt_str = "<" + str(array) + cpp_typename[data_type]
            size = struct.calcsize(fmt_str)  # 目标数据大小
            buff = ctypes.create_string_buffer(size)  # 目标数据缓冲
            success = ReadProcessMemory(pvz_handle, offset, ctypes.byref(buff), size, ctypes.byref(bytes_read))
            if success == 0 or bytes_read.value != size:
                critical("读取内存失败, 错误代码 %d." % GetLastError())

            result = struct.unpack(fmt_str, buff.raw)

    memory_lock.release()

    debug("读取内存, 类型 %s, 地址 %s, 数量 %d, 结果 %s." % (data_type, str(address), array, str(result)))
    if array == 1:
        return result[0]
    else:
        return result


def write_memory(data_type, values, *address):
    """
    写入内存数据.

    @参数 data_type(str): 数据类型, 取自 C/C++ 语言关键字, 可选值 ["char", "bool", "unsigned char", "short", "unsigned short", "int", "unsigned int", "long", "unsigned long", "long long", "unsigned long long", "float", "double"].

    @参数 values(int/float/bool/list/tuple): 需要写入的数据, 多个数据采用列表或者元组形式.

    @参数 address(int): 地址, 可为多级偏移.

    @示例:

    >>> WriteMemory("int", 8000, 0x6a9ec0, 0x768, 0x5560)

    >>> WriteMemory("unsigned char", [0xb0, 0x01, 0xc3], 0x0041d7d0)
    """

    if not is_valid():
        critical("目标进程不可用, 写内存失败.")

    # 将单个数据转换为列表方便统一处理
    if not isinstance(values, (tuple, list)):
        values = [values]

    memory_lock.acquire()

    level = len(address)  # 偏移级数
    offset = ctypes.c_void_p()  # 内存地址
    buffer = ctypes.c_uint()  # 中间数据缓冲
    bytes_read = ctypes.c_ulong()  # 已读字节数
    bytes_written = ctypes.c_ulong()  # 已写字节数

    for i in range(level):
        offset.value = buffer.value + address[i]

        if i != level - 1:
            size = ctypes.sizeof(buffer)
            success = ReadProcessMemory(pvz_handle, offset, ctypes.byref(buffer), size, ctypes.byref(bytes_read))
            if success == 0 or bytes_read.value != size:
                critical("读取内存失败, 错误代码 %d." % GetLastError())

        else:
            array = len(values)  # 目标数据的数量
            fmt_str = "<" + str(array) + cpp_typename[data_type]
            size = struct.calcsize(fmt_str)  # 目标数据大小
            buff = ctypes.create_string_buffer(size)  # 创建目标数据缓冲
            buff.value = struct.pack(fmt_str, *values)  # 将目标数据载入缓冲区
            success = WriteProcessMemory(pvz_handle, offset, ctypes.byref(buff), size, ctypes.byref(bytes_written))
            if success == 0 or bytes_written.value != size:
                critical("写入内存失败, 错误代码 %d." % GetLastError())

    memory_lock.release()

    debug("写入内存, 类型 %s, 数值 %s, 地址 %s." % (data_type, str(values), str(address)))


### 窗口置顶


def active_pvz():
    """
    激活 PvZ 窗口. (后台时不激活.)
    """
    if pvz_hwnd.value is not None:
        SetActiveWindow(pvz_hwnd)


def set_pvz_foreground():
    """
    激活并前台显示 PvZ 窗口.
    """
    if pvz_hwnd.value is not None:
        SetForegroundWindow(pvz_hwnd)


def set_pvz_top_most(on=True):
    """
    置顶显示游戏窗口.

    @参数 on(bool): 是否开启.
    """
    if pvz_hwnd.value is not None:
        SetWindowPos(
            pvz_hwnd,  #
            HWND_TOPMOST if on else HWND_NOTOPMOST,  #
            0,  #
            0,  #
            0,  #
            0,  #
            SWP_NOMOVE | SWP_NOSIZE | SWP_SHOWWINDOW,  #
        )


### 进程优先级


def set_pvz_high_priority():
    """
    游戏进程以高优先级运行. 可在一定程度上提高帧率.
    """
    if GetPriorityClass(pvz_handle) != REALTIME_PRIORITY_CLASS:
        SetPriorityClass(pvz_handle, HIGH_PRIORITY_CLASS)


# 汇编代码

asm_code = bytes()


# 初始化
def asm_init():
    global asm_code
    asm_code = bytes()


# unsigned char
def asm_add_byte(code):
    global asm_code
    asm_code += struct.pack("<1B", code)


# unsigned short
def asm_add_word(code):
    global asm_code
    asm_code += struct.pack("<1H", code)


# unsigned int
def asm_add_dword(code):
    global asm_code
    asm_code += struct.pack("<1I", code)


# bytes from list/tuple
def asm_add_bytes(codes):
    for code in codes:
        asm_add_byte(code)


# push 0x12345678
def asm_push(code):
    asm_add_byte(0x68)
    asm_add_dword(code)


# mov exx, 0x12345678
asm_mov_exx_code = {
    "eax": [0xB8],
    "ebx": [0xBB],
    "ecx": [0xB9],
    "edx": [0xBA],
    "esi": [0xBE],
    "edi": [0xBF],
    "ebp": [0xBD],
    "esp": [0xBC]
}


def asm_mov_exx(register, code):
    asm_add_bytes(asm_mov_exx_code[register])
    asm_add_dword(code)


# add exx, 0x12345678
asm_add_exx_code = {
    "eax": [0x05],
    "ebx": [0x81, 0xC3],
    "ecx": [0x81, 0xC1],
    "edx": [0x81, 0xC2],
    "esi": [0x81, 0xC6],
    "edi": [0x81, 0xC7],
    "ebp": [0x81, 0xC5],
    "esp": [0x81, 0xC4],
}


def asm_add_exx(register, code):
    asm_add_bytes(asm_add_exx_code[register])
    asm_add_dword(code)


# mov exx, ds:[0x12345678]
asm_mov_exx_dword_ptr_code = {
    "eax": [0x3E, 0xA1],
    "ebx": [0x3E, 0x8B, 0x1D],
    "ecx": [0x3E, 0x8B, 0x0D],
    "edx": [0x3E, 0x8B, 0x15],
    "esi": [0x3E, 0x8B, 0x35],
    "edi": [0x3E, 0x8B, 0x3D],
    "ebp": [0x3E, 0x8B, 0x2D],
    "esp": [0x3E, 0x8B, 0x25],
}


def asm_mov_exx_dword_ptr(register, code):
    asm_add_bytes(asm_mov_exx_dword_ptr_code[register])
    asm_add_dword(code)


# mov exx, [exx + 0x12345678]
asm_mov_exx_dword_ptr_exx_add_code = {
    "eax": [0x8B, 0x80],
    "ebx": [0x8B, 0x9B],
    "ecx": [0x8B, 0x89],
    "edx": [0x8B, 0x92],
    "esi": [0x8B, 0xB6],
    "edi": [0x8B, 0xBF],
    "ebp": [0x8B, 0xAD],
    "esp": [0x8B, 0xA4, 0x24],
}


def asm_mov_exx_dword_ptr_exx_add(register, code):
    asm_add_bytes(asm_mov_exx_dword_ptr_exx_add_code[register])
    asm_add_dword(code)


# push exx
asm_push_exx_code = {
    "eax": [0x50],
    "ebx": [0x53],
    "ecx": [0x51],
    "edx": [0x52],
    "esi": [0x56],
    "edi": [0x57],
    "ebp": [0x55],
    "esp": [0x54]
}


def asm_push_exx(register):
    asm_add_bytes(asm_push_exx_code[register])


# pop exx
asm_pop_exx_code = {
    "eax": [0x58],
    "ebx": [0x5B],
    "ecx": [0x59],
    "edx": [0x5A],
    "esi": [0x5E],
    "edi": [0x5F],
    "ebp": [0x5D],
    "esp": [0x5C]
}


def asm_pop_exx(register):
    asm_add_bytes(asm_pop_exx_code[register])


# ret
def asm_ret():
    asm_add_byte(0xC3)


# call 0x12345678
# call $ + 7
# jmp short $ + 8
# push 0x12345678
# ret
def asm_call(address):
    asm_add_bytes([0xE8, 0x02, 0x00, 0x00, 0x00])
    asm_add_bytes([0xEB, 0x06])
    asm_push(address)
    asm_ret()


# inject
def asm_code_inject():
    """
    远程注入汇编代码.
    """

    size = len(asm_code)

    # thread_addr = LPVOID()
    thread_addr = VirtualAllocEx(pvz_handle, None, size, MEM_COMMIT, PAGE_EXECUTE_READWRITE)

    if thread_addr is not None:

        bytes_written = ctypes.c_ulong()
        success = WriteProcessMemory(pvz_handle, thread_addr, asm_code, size, ctypes.byref(bytes_written))
        if success == 0 or bytes_written.value != size:
            critical("写入汇编代码失败, 错误代码 %d." % GetLastError())

        # thread_handle = HANDLE()
        start = LPTHREAD_START_ROUTINE(thread_addr)
        thread_handle = CreateRemoteThread(pvz_handle, None, 0, start, None, 0, None)

        if thread_handle is not None:
            success = WaitForSingleObject(thread_handle, INFINITE)
            if success == WAIT_FAILED:
                critical("等待对象返回失败, 错误代码 %d." % GetLastError())
            success = CloseHandle(thread_handle)
            if success == 0:
                critical("关闭远程线程句柄失败, 错误代码 %d." % GetLastError())
        else:
            critical("创建远程线程失败, 错误代码 %d." % GetLastError())

        success = VirtualFreeEx(pvz_handle, thread_addr, 0, MEM_RELEASE)
        if success == 0:
            critical("释放内存失败, 错误代码 %d." % GetLastError())

    else:
        critical("分配内存失败, 错误代码 %d." % GetLastError())

    info("远程注入代码: %s." % str([hex(x) for x in asm_code]))


# 避免崩溃
def asm_code_inject_safely():
    if pvz_ver() == "1.0.0.1051":
        write_memory("unsigned char", 0xFE, 0x00552014)
    else:
        write_memory("unsigned char", 0xFE, 0x00552244)
    time.sleep(0.01)
    if is_valid():
        asm_code_inject()
    if pvz_ver() == "1.0.0.1051":
        write_memory("unsigned char", 0xDB, 0x00552014)
    else:
        write_memory("unsigned char", 0xDB, 0x00552244)


### 键盘操作


def press_esc():
    """
    敲击 退出 键.
    """
    PostMessageW(pvz_hwnd, WM_KEYDOWN, VK_ESCAPE, 0)
    PostMessageW(pvz_hwnd, WM_KEYUP, VK_ESCAPE, 0)


def press_space():
    """
    敲击 空格 键.
    """
    PostMessageW(pvz_hwnd, WM_KEYDOWN, VK_SPACE, 0)
    PostMessageW(pvz_hwnd, WM_KEYUP, VK_SPACE, 0)


def press_enter():
    """
    敲击 回车 键.
    """
    PostMessageW(pvz_hwnd, WM_KEYDOWN, VK_RETURN, 0)
    PostMessageW(pvz_hwnd, WM_KEYUP, VK_RETURN, 0)


def press_left():
    """
    敲击 左方向 键.
    """
    PostMessageW(pvz_hwnd, WM_KEYDOWN, VK_LEFT, 0)
    PostMessageW(pvz_hwnd, WM_KEYUP, VK_LEFT, 0)


def press_up():
    """
    敲击 上方向 键.
    """
    PostMessageW(pvz_hwnd, WM_KEYDOWN, VK_UP, 0)
    PostMessageW(pvz_hwnd, WM_KEYUP, VK_UP, 0)


def press_right():
    """
    敲击 右方向 键.
    """
    PostMessageW(pvz_hwnd, WM_KEYDOWN, VK_RIGHT, 0)
    PostMessageW(pvz_hwnd, WM_KEYUP, VK_RIGHT, 0)


def press_down():
    """
    敲击 下方向 键.
    """
    PostMessageW(pvz_hwnd, WM_KEYDOWN, VK_DOWN, 0)
    PostMessageW(pvz_hwnd, WM_KEYUP, VK_DOWN, 0)


def press_key(key):
    """
    敲击按键. 可选值 '0' - '9' 'A' - 'Z'.
    """
    code = ord(key)
    PostMessageW(pvz_hwnd, WM_KEYDOWN, code, 0)
    PostMessageW(pvz_hwnd, WM_KEYUP, code, 0)


def press_keys(keys):
    """
    敲击一系列按键.

    @参数 keys(str): 按键字符串, 由 '0' - '9' 'A' - 'Z' 组成.

    @示例:

    >>> PressKeys("FUTURE")  # 智慧树指令, 使僵尸带上眼镜
    """
    for k in keys:
        press_key(k.upper())


### 鼠标操作

dpi_scale = 1.0


def get_dpi_scale():
    """
    获取 DPI 缩放比例.
    """
    screen = GetDC(None)
    if screen is not None:
        virtual_width = GetDeviceCaps(screen, HORZRES)
        physical_width = GetDeviceCaps(screen, DESKTOPHORZRES)
        ReleaseDC(None, screen)
        scale = physical_width / virtual_width
    else:
        scale = 1.0

    global dpi_scale
    dpi_scale = scale
    info("获取 DPI 缩放比例: %s." % dpi_scale)


def set_dpi_scale(scale):
    """
    设置 DPI 缩放比例.

    @参数 scale(float): 比例系数.
    """
    global dpi_scale
    dpi_scale = scale
    info("设置 DPI 缩放比例: %s." % dpi_scale)


def MAKELONG(low, high):
    if dpi_scale != 1.0:
        low, high = int(low / dpi_scale), int(high / dpi_scale)
    else:
        low, high = int(low), int(high)
    return ((high & 0xFFFF) << 16) | (low & 0xFFFF)


def left_down(x, y):
    """
    鼠标左键按下.
    """
    coord = MAKELONG(x, y)
    PostMessageW(pvz_hwnd, WM_LBUTTONDOWN, MK_LBUTTON, coord)


def left_up(x, y):
    """
    鼠标左键弹起.
    """
    coord = MAKELONG(x, y)
    PostMessageW(pvz_hwnd, WM_LBUTTONUP, MK_LBUTTON, coord)


def left_click(x, y):
    """
    鼠标左键单击.

    @参数 x(int): 横坐标, 单位像素. 建议范围 [0, 799].

    @参数 y(int): 纵坐标, 单位像素. 建议范围 [0, 599].

    @示例:

    >>> LeftClick(108, 42)  # 左键单击卡槽第一张卡片的位置
    """
    coord = MAKELONG(x, y)
    PostMessageW(pvz_hwnd, WM_LBUTTONDOWN, MK_LBUTTON, coord)
    PostMessageW(pvz_hwnd, WM_LBUTTONUP, MK_LBUTTON, coord)


def right_down(x, y):
    """
    鼠标右键按下.
    """
    coord = MAKELONG(x, y)
    PostMessageW(pvz_hwnd, WM_RBUTTONDOWN, MK_RBUTTON, coord)


def right_up(x, y):
    """
    鼠标右键弹起.
    """
    coord = MAKELONG(x, y)
    PostMessageW(pvz_hwnd, WM_RBUTTONUP, MK_RBUTTON, coord)


def right_click(x, y):
    """
    鼠标右键单击.

    @参数 x(int): 横坐标, 单位像素. 建议范围 [0, 799].

    @参数 y(int): 纵坐标, 单位像素. 建议范围 [0, 599].

    @示例:

    >>> RightClick(399, 299)  # 右键单击场地中间位置
    """
    coord = MAKELONG(x, y)
    PostMessageW(pvz_hwnd, WM_RBUTTONDOWN, MK_RBUTTON, coord)
    PostMessageW(pvz_hwnd, WM_RBUTTONUP, MK_RBUTTON, coord)


def special_button_click(x, y):
    """
    适用于模仿者按钮和菜单按钮的特殊点击.

    调用的时候不要把鼠标光标放在游戏窗口内.

    @参数 x(int): 横坐标, 单位像素. 建议范围 [0, 799].

    @参数 y(int): 纵坐标, 单位像素. 建议范围 [0, 599].

    @示例:

    >>> ButtonClick(490, 550)  # 选卡界面点击模仿者卡片

    >>> ButtonClick(740, 10)  # 点击菜单按钮
    """
    left_down(x, y)
    right_down(x, y)
    left_up(x, y)
    right_up(x, y)
    time.sleep(0.01)


### 卡片和僵尸名称

seeds_string = [
    ["Peashooter", "豌豆射手", "豌豆", "单发"],
    ["Sunflower", "向日葵", "小向", "太阳花", "花"],
    ["Cherry Bomb", "樱桃炸弹", "樱桃", "炸弹", "爆炸", "草莓", "樱"],
    ["Wall-nut", "坚果", "坚果墙", "墙果", "建国", "柠檬圆"],
    ["Potato Mine", "土豆地雷", "土豆", "地雷", "土豆雷"],
    ["Snow Pea", "寒冰射手", "冰豆", "冰冻豌豆", "冰豌豆", "雪花豌豆", "雪花"],
    ["Chomper", "大嘴花", "大嘴", "食人花", "咀嚼者", "食"],
    ["Repeater", "双重射手", "双发射手", "双重", "双发", "双发豌豆"],
    ["Puff-shroom", "小喷菇", "小喷", "喷汽蘑菇", "烟雾蘑菇", "免费蘑菇", "炮灰菇", "小蘑菇", "免费货", "免费", "紫蘑菇"],
    ["Sun-shroom", "阳光菇", "阳光", "阳光蘑菇"],
    ["Fume-shroom", "大喷菇", "大喷", "烟雾喷菇", "大蘑菇", "喷子", "喷"],
    ["Grave Buster", "咬咬碑", "墓碑吞噬者", "墓碑破坏者", "噬碑藤", "墓碑", "墓碑苔藓", "苔藓"],
    ["Hypno-shroom", "迷糊菇", "魅惑菇", "魅惑", "迷惑菇", "迷蘑菇", "催眠蘑菇", "催眠", "花蘑菇", "毒蘑菇"],
    ["Scaredy-shroom", "胆小菇", "胆小", "胆怯蘑菇", "胆小鬼蘑菇", "杠子蘑菇"],
    ["Ice-shroom", "冰川菇", "寒冰菇", "冰菇", "冷冻蘑菇", "冰蘑菇", "冰莲菇", "面瘫", "蓝冰", "原版冰", "冰"],
    ["Doom-shroom", "末日菇", "毁灭菇", "核蘑菇", "核弹", "核武", "毁灭", "末日蘑菇", "末日", "黑核", "原版核", "核"],
    ["Lily Pad", "莲叶", "睡莲", "荷叶", "莲"],
    ["Squash", "窝瓜", "倭瓜", "窝瓜大叔", "倭瓜大叔", "镇压者"],
    ["Threepeater", "三重射手", "三线射手", "三头豌豆", "三联装豌豆", "三重", "三线", "三头", "三管", "管"],
    ["Tangle Kelp", "缠绕水草", "缠绕海草", "缠绕海藻", "缠绕海带", "水草", "海草", "海藻", "海带", "马尾藻", "绿毛线", "毛线"],
    ["Jalapeno", "火爆辣椒", "辣椒", "墨西哥胡椒", "墨西哥辣椒", "辣", "椒"],
    ["Spikeweed", "地刺", "刺", "尖刺", "尖刺杂草", "棘草", "荆棘"],
    ["Torchwood", "火炬树桩", "火树桩", "火树", "火炬", "树桩", "火炬木", "火"],
    ["Tall-nut", "高坚果", "搞基果", "高建国", "巨大墙果", "巨大", "高墙果", "大墙果", "大土豆"],
    ["Sea-shroom", "水兵菇", "海蘑菇"],
    ["Plantern", "路灯花", "灯笼", "路灯", "灯笼草", "灯笼花", "吐槽灯", "灯"],
    ["Cactus", "仙人掌", "小仙", "掌"],
    ["Blover", "三叶草", "三叶", "风扇", "吹风", "愤青"],
    ["Split Pea", "双向射手", "裂荚射手", "裂荚", "双头", "分裂豌豆", "双头豌豆"],
    ["Starfruit", "星星果", "杨桃", "星星", "五角星", "五星黄果", "1437", "大帝", "桃"],
    ["Pumpkin", "南瓜头", "南瓜", "南瓜罩", "南瓜壳", "套"],
    ["Magnet-shroom", "磁力菇", "磁铁", "磁力蘑菇", "磁"],
    ["Cabbage-pult", "卷心菜投手", "包菜", "卷心菜", "卷心菜投抛者", "包菜投掷手"],
    ["Flower Pot", "花盆", "盆"],
    ["Kernel-pult", "玉米投手", "玉米", "黄油投手", "玉米投抛者", "玉米投掷手"],
    ["Coffee Bean", "咖啡豆", "咖啡", "兴奋剂", "春药"],
    ["Garlic", "大蒜", "蒜"],
    ["Umbrella Leaf", "萝卜伞", "叶子保护伞", "伞型保护叶", "莴苣", "萝卜", "白菜", "保护伞", "叶子伞", "伞叶", "叶子", "伞", "叶"],
    ["Marigold", "金盏花", "金盏草", "金盏菊", "吐钱花"],
    ["Melon-pult", "西瓜投手", "西瓜", "绿皮瓜", "瓜", "西瓜投抛者", "西瓜投掷手"],
    ["Gatling Pea", "机枪射手", "加特林豌豆", "格林豌豆", "加特林", "机枪", "枪"],
    ["Twin Sunflower", "双胞向日葵", "双子向日葵", "双头葵花", "双胞", "双子", "双向", "双花"],
    ["Gloom-shroom", "多嘴小蘑菇", "忧郁蘑菇", "忧郁", "忧郁菇", "章鱼", "曾哥", "曾哥蘑菇", "曾"],
    ["Cattail", "猫尾草", "香蒲", "猫尾", "猫尾香蒲", "小猫香蒲", "小猫", "猫"],
    ["Winter Melon", "冰西瓜", "'冰'瓜", '"冰"瓜', "冰瓜", "冰冻西瓜", "冬季西瓜"],
    ["Gold Magnet", "吸金菇", "吸金磁", "吸金草", "金磁铁", "吸金", "磁力金钱菇"],
    ["Spikerock", "钢地刺", "钢刺", "地刺王", "尖刺岩石", "尖刺石", "石荆棘"],
    ["Cob Cannon", "玉米加农炮", "玉米炮", "加农炮", "春哥", "春哥炮", "炮", "春", "神"],
]

# # 确保没有重复项, 发布时注释掉
# seeds_string_all = []
# for items in seeds_string:
#     seeds_string_all += items
# print(len(seeds_string_all))
# assert len(seeds_string_all) == len(set(seeds_string_all))

# 模仿者卡片前缀
seeds_imitater_string = ["Imitater", "imitater", "变身茄子", "模仿者", "模仿", "复制", "白", "小白", "克隆"]

# 整理成字典方便快速查找
# key: 卡片名称
# value: 卡片代号 0~47 (模仿者 +48)
seeds_string_dict = {}
for i, items in enumerate(seeds_string):
    for item in items:
        # 绿卡 紫卡
        seeds_string_dict[item] = i
        # 白卡
        for j, im in enumerate(seeds_imitater_string):
            seeds_string_dict[im + item] = i + 48
            seeds_string_dict[im + " " + item] = i + 48
# print(len(seeds_string_dict))  # it's huge!!!

zombies_string = [
    ["Zombie", "普僵", "普通", "领带"],
    ["Flag Zombie", "旗帜", "摇旗", "旗子"],
    ["Conehead Zombie", "路障"],
    ["Pole Vaulting Zombie", "撑杆", "撑杆跳"],
    ["Buckethead Zombie", "铁桶"],
    ["Newspaper Zombie", "读报", "报纸"],
    ["Screen Door Zombie", "铁门", "铁栅门", "门板"],
    ["Football Zombie", "橄榄", "橄榄球"],
    ["Dancing Zombie", "舞王", "MJ"],
    ["Backup Dancer", "伴舞", "舞伴"],
    ["Ducky Tube Zombie", "鸭子", "救生圈"],
    ["Snorkel Zombie", "潜水"],
    ["Zomboni", "冰车", "制冰车"],
    ["Zombie Bobsled Team", "雪橇", "雪橇队", "雪橇小队"],
    ["Dolphin Rider Zombie", "海豚", "海豚骑士"],
    ["Jack-in-the-Box Zombie", "小丑", "玩偶匣"],
    ["Balloon Zombie", "气球"],
    ["Digger Zombie", "矿工", "挖地"],
    ["Pogo Zombie", "跳跳", "弹跳"],
    ["Zombie Yeti", "雪人"],
    ["Bungee Zombie", "蹦极", "小偷"],
    ["Ladder Zombie", "扶梯", "梯子"],
    ["Catapult Zombie", "投篮", "投篮车", "篮球"],
    ["Gargantuar", "白眼", "伽刚特尔", "巨人"],
    ["Imp", "小鬼", "小恶魔", "IMP"],
    ["Dr. Zomboss", "僵王", "僵博"],
    ["Peashooter Zombie", "豌豆"],
    ["Wall-nut Zombie", "坚果"],
    ["Jalapeno Zombie", "辣椒"],
    ["Gatling Pea Zombie", "机枪", "加特林"],
    ["Squash Zombie", "倭瓜", "窝瓜"],
    ["Tall-nut Zombie", "高坚果"],
    ["GigaGargantuar", "红眼", "暴走伽刚特尔", "红眼巨人"],
]

zombies_string_dict = {}
for i, zombies in enumerate(zombies_string):
    for j, z in enumerate(zombies):
        if j == 0:  # 第一个英文名称
            zombies_string_dict[z] = i
        else:
            zombies_string_dict[z] = i
            zombies_string_dict[z + "僵尸"] = i
# print(len(zombies_string_dict))

### 延时


def thread_sleep_for(time_cs):
    """
    线程睡眠延时.

    实际睡眠时间依赖于操作系统线程切换时间片精度.

    @参数 time_cs(float): 时间, 单位 cs, 精度 0.1.
    """

    if time_cs > 0.0:
        time.sleep(time_cs / 100)
    elif time_cs == 0.0:
        pass
    else:
        error("线程睡眠时间不能小于零.")


# 细微延时
def delay_a_little_time():
    thread_sleep_for(0.1)  # 1ms


### 子线程装饰器


def running_in_thread(func):
    """
    将此装饰器应用到需要在子线程运行的函数上.

    定义一个函数, 应用该装饰器, 则函数在调用的时候会运行在单独的线程中.

    @示例:

    >>> @RunningInThread
    >>> def func():
    >>>     pass
    >>> # ...
    >>> func()
    """
    @functools.wraps(func)  # 复制原函数元信息
    def wrapper(*args, **kwargs):
        thread = threading.Thread(target=func, args=args, kwargs=kwargs)
        # thread.setDaemon(True)  # 守护线程
        thread.start()

    return wrapper



# coding=utf-8
"""
Python vs. Zombies
"""

__name__ = "pvz"
__version__ = "4.0.1"
__description__ = "Python vs. Zombies"
__date__ = "2020-08-29"
__status__ = "Production"
__author__ = "lmintlcx"
__copyright__ = "Copyright 2018-2020, lmintlcx"
__credits__ = ["no_doudle", "a418569882"]
__license__ = "GPL"
__maintainer__ = "lmintlcx"
__email__ = "lmintlcx@gmail.com"

import platform
import sys
import gc
import atexit

### 检查操作系统和 Python 版本

if platform.system() != "Windows":
    raise Exception("本项目 (pvz) 只支持在 Windows 系统上使用.")

if sys.hexversion < 0x030400f0:
    raise Exception("本项目 (pvz) 要求版本号 >=3.4.0 的 Python 运行环境.")

print("当前版本: %s" % __version__)
print("在线教程: https://pvz.lmintlcx.com/scripts/")

from .core import *
from .extra import *

### documented api

## 读写内存
from .core import read_memory as ReadMemory
from .core import write_memory as WriteMemory

## 模拟鼠标点击
from .core import left_click as LeftClick
from .core import right_click as RightClick
from .core import special_button_click as ButtonClick

## 模拟键盘敲击
from .core import press_esc as PressEsc
from .core import press_space as PressSpace
from .core import press_enter as PressEnter
from .core import press_keys as PressKeys

## 功能修改
from .extra import set_zombies as SetZombies

## 选卡/更新炮列表
from .extra import select_seeds_and_lets_rock as SelectCards
from .extra import update_cob_cannon_list as UpdatePaoList

## 阻塞延时
from .core import thread_sleep_for as Sleep
from .extra import game_delay_for as Delay
from .extra import until_relative_time_after_refresh as Prejudge
from .extra import until_relative_time as Until

## 场地点击
from .extra import get_mouse_lock as MouseLock
from .extra import safe_click as SafeClick
from .extra import click_seed as ClickSeed
from .extra import click_shovel as ClickShovel
from .extra import click_grid as ClickGrid

## 主要操作
from .extra import use_seed as Card
from .extra import use_shovel as Shovel
from .extra import fire_cob as Pao
from .extra import fire_cob_on_roof as RoofPao
from .extra import skip_cob_index as Skip

## 子线程操作
from .extra import running_in_thread as RunningInThread
from .extra import auto_collect as AutoCollect
from .extra import auto_fill_ice as IceSpots
from .extra import activate_ice as Coffee

__all__ = [
    # 读写内存
    "ReadMemory",
    "WriteMemory",
    # 模拟鼠标点击
    "LeftClick",
    "RightClick",
    "ButtonClick",
    # 模拟键盘敲击
    "PressEsc",
    "PressSpace",
    "PressEnter",
    "PressKeys",
    # 功能修改
    "SetZombies",
    # 选卡/更新炮列表
    "SelectCards",
    "UpdatePaoList",
    # 阻塞延时
    "Sleep",
    "Delay",
    "Prejudge",
    "Until",
    # 场地点击
    "MouseLock",
    "SafeClick",
    "ClickSeed",
    "ClickShovel",
    "ClickGrid",
    # 主要操作
    "Card",
    "Shovel",
    "Pao",
    "RoofPao",
    "Skip",
    # 子线程操作
    "RunningInThread",
    "AutoCollect",
    "IceSpots",
    "Coffee",
]

### 启动和退出时的特殊处理


def _on_start():
    # 其实不需要, 游戏启动后会变成 0.5ms
    timeBeginPeriod(1)  # res == TIMERR_NOERROR

    # 时间敏感
    gc.disable()
    sys.setswitchinterval(0.001)

    SetPriorityClass(GetCurrentProcess(), HIGH_PRIORITY_CLASS)

    get_dpi_scale()
    # set_dpi_scale(1.25)  # 自动获取缩放率出错则手动设置

    if find_pvz():
        ui = game_ui()
        if ui in (2, 3):
            set_pvz_foreground()
            set_pvz_high_priority()
            update_game_scene()
            update_seeds_list() if ui == 3 else None
            update_cob_cannon_list()
    else:
        critical("游戏未开启或者游戏版本不受支持!")

    enable_logging(False)  # 是否输出调试日志
    set_logging_level("INFO")


def _on_exit():
    timeEndPeriod(1)  # res == TIMERR_NOERROR

    if is_valid():
        CloseHandle(pvz_handle)


_on_start()
atexit.register(_on_exit)



# coding=utf-8

from distutils.core import setup

setup(
    name="pvz",
    version="4.0.1",
    author="lmintlcx",
    author_email="lmintlcx@gmail.com",
    url="https://github.com/lmintlcx/pvzscripts",
    license="GPL",
    description="Python vs. Zombies",
    long_description="Plants vs. Zombies TAS Framework",
    platforms=["Windows"],
    python_requires=">=3.4.0",
    packages=["pvz"],
)



# coding=utf-8

from .core import *

### 读取常用信息


def game_on():
    """
    @返回值 (bool): 游戏是否开启, 没开则会尝试重新查找一次.
    """
    if is_valid():
        return True
    else:
        return find_pvz()


def game_ui():
    """
    @返回值 (int): 游戏界面.

    1: 主界面, 2: 选卡, 3: 正常游戏/战斗, 4: 僵尸进屋, 7: 模式选择.
    """
    return read_memory("int", 0x6A9EC0, 0x7FC)


def game_mode():
    """
    @返回值 (int): 游戏模式, 13 为生存无尽.
    """
    return read_memory("int", 0x6A9EC0, 0x7F8)


def game_scene():
    """
    @返回值 (int): 游戏场景/场地/地图.

    0: 白天, 1: 黑夜, 2: 泳池, 3: 浓雾, 4: 屋顶, 5: 月夜, 6: 蘑菇园, 7: 禅境花园, 8: 水族馆, 9: 智慧树.
    """
    return read_memory("int", 0x6A9EC0, 0x768, 0x554C)


def game_paused():
    """
    @返回值 (bool): 当前游戏是否暂停.
    """
    return read_memory("bool", 0x6A9EC0, 0x768, 0x164)


def mouse_in_game():
    """
    @返回值 (bool): 鼠标是否在游戏窗口内部.
    """
    return read_memory("bool", 0x6A9EC0, 0x768, 0x138, 0x18)  # 0x6A9EC0, 0x768, 0x59


def mouse_have_something():
    """
    @返回值 (bool): 鼠标是否选中卡炮或铲子.
    """
    return read_memory("int", 0x6A9EC0, 0x768, 0x138, 0x30) in (1, 6, 8)


def game_clock():
    """
    @返回值 (int): 内部时钟, 游戏暂停和选卡时会暂停计时.
    """
    return read_memory("int", 0x6A9EC0, 0x768, 0x5568)


def wave_init_countdown():
    """
    @返回值 (int): 刷新倒计时初始值.
    """
    return read_memory("int", 0x6A9EC0, 0x768, 0x55A0)


def wave_countdown():
    """
    @返回值 (int): 下一波刷新倒计时, 触发刷新时重置为 200, 减少至 1 后刷出下一波.
    """
    return read_memory("int", 0x6A9EC0, 0x768, 0x559C)


def huge_wave_countdown():
    """
    @返回值 (int): 大波刷新倒计时, 对于旗帜波, 刷新倒计时减少至 4 后停滞, 由该值代替减少.
    """
    return read_memory("int", 0x6A9EC0, 0x768, 0x55A4)


def current_wave():
    """
    @返回值 (int): 已刷新波数.
    """
    return read_memory("int", 0x6A9EC0, 0x768, 0x557C)


### 修改出怪


# 从出怪种子生成出怪类型
def update_zombies_type():
    write_memory("bool", [False] * 33, 0x6A9EC0, 0x768, 0x54D4)
    asm_init()
    asm_mov_exx_dword_ptr("esi", 0x6A9EC0)
    asm_mov_exx_dword_ptr_exx_add("esi", 0x768)
    asm_mov_exx_dword_ptr_exx_add("esi", 0x160)
    asm_call(0x00425840) if pvz_ver() == "1.0.0.1051" else asm_call(0x004258A0)
    asm_ret()
    asm_code_inject_safely()


# 从出怪类型生成出怪列表
def update_zombies_list():
    asm_init()
    asm_mov_exx_dword_ptr("edi", 0x6A9EC0)
    asm_mov_exx_dword_ptr_exx_add("edi", 0x768)
    asm_call(0x004092E0) if pvz_ver() == "1.0.0.1051" else asm_call(0x004092F0)
    asm_ret()
    asm_code_inject_safely()


# 更新选卡界面出怪预览
def update_zombies_preview():
    write_memory("byte", 0x80, 0x0043A153) if pvz_ver() == "1.0.0.1051" else write_memory("byte", 0x80, 0x0043A1C3)
    asm_init()
    asm_mov_exx_dword_ptr("ebx", 0x6A9EC0)
    asm_mov_exx_dword_ptr_exx_add("ebx", 0x768)
    asm_call(0x0040DF70) if pvz_ver() == "1.0.0.1051" else asm_call(0x0040DF80)
    asm_mov_exx_dword_ptr("eax", 0x6A9EC0)
    asm_mov_exx_dword_ptr_exx_add("eax", 0x768)
    asm_mov_exx_dword_ptr_exx_add("eax", 0x15C)
    asm_push_exx("eax")
    asm_call(0x0043A140) if pvz_ver() == "1.0.0.1051" else asm_call(0x0043A1B0)
    asm_ret()
    asm_code_inject_safely()
    write_memory("byte", 0x85, 0x0043A153) if pvz_ver() == "1.0.0.1051" else write_memory("byte", 0x85, 0x0043A1C3)


# 僵尸名称统一转换为代号表示
def zombie_name_to_index(zombie):
    if isinstance(zombie, str):
        if zombie in zombies_string_dict:
            return zombies_string_dict[zombie]
        else:
            error("未知僵尸名称: %s." % zombie)
    else:  # int
        if zombie in range(33):
            return zombie
        else:
            error("未知僵尸类型代号: %s." % zombie)


def set_internal_spawn(zombies=None):
    """
    内置刷怪, 由游戏自带函数生成出怪列表. 普僵不管有没有选都是必出.

    @参数 zombies(list[str/int]): 包含僵尸名称或代号的列表.
    """

    if not (game_on() and game_ui() in (2, 3)):
        return

    if zombies is None:
        zombies = []
    zombies += [0]  # 普僵必出
    zombies = [zombie_name_to_index(z) for z in zombies]
    zombies = list(set(zombies))

    zombies_type_offset = read_memory("unsigned int", 0x6A9EC0, 0x768) + 0x54D4
    write_memory("bool", [True if i in zombies else False for i in range(33)], zombies_type_offset)
    update_zombies_list()

    if game_ui() == 2:
        update_zombies_preview()


def set_customize_spawn(zombies=None):
    """
    自定义刷怪, 由脚本生成并填充出怪列表.

    @参数 zombies(list[str/int]): 包含僵尸名称或代号的列表.
    """

    if not (game_on() and game_ui() in (2, 3)):
        return

    if zombies is None:
        zombies = [0]  # 默认普僵
    zombies = [zombie_name_to_index(z) for z in zombies]
    zombies = list(set(zombies))

    zombies_list = [0] * 1000

    has_flag = 1 in zombies
    has_yeti = 19 in zombies
    has_bungee = 20 in zombies
    limit_flag = True
    limit_yeti = True
    limit_bungee = True

    count = 0
    for i in range(33):
        if i in zombies:
            count += 1

    if count > 0:

        zombie_type = 0
        for i in range(1000):
            while True:
                zombie_type += 1
                zombie_type %= 33
                if not (  #
                    (zombie_type not in zombies)  #
                        or (has_flag and limit_flag and zombie_type == 1)  #
                        or (has_yeti and limit_yeti and zombie_type == 19)  #
                        or (has_bungee and limit_bungee and zombie_type == 20)  #
                ):  #
                    break
            zombies_list[i] = zombie_type

        index_flag = [450, 950]
        index_bungee = [451, 452, 453, 454, 951, 952, 953, 954]

        if has_flag and limit_flag:
            for i in index_flag:
                zombies_list[i] = 1

        if has_bungee and limit_bungee:
            for i in index_bungee:
                zombies_list[i] = 20

        if has_yeti and limit_yeti:
            i = 0
            while True:
                i = random.randint(0, 999)
                if not ((i in index_flag) or (i in index_bungee)):
                    break
            zombies_list[i] = 19

    write_memory("unsigned int", zombies_list, 0x6A9EC0, 0x768, 0x6B4)

    if game_ui() == 2:
        update_zombies_preview()


def set_zombies(zombies=None, mode="极限刷怪"):
    """
    设置出怪.

    旗帜只会在每个大波出现一只, 雪人只会出现一只, 蹦极只会在大波出现.

    @参数 zombies(list[str/int]): 包含僵尸名称或代号的列表, 建议 8~12 种.

    @参数 mode(str): 刷怪模式, 默认使用极限刷怪. 可选值 "自然刷怪" "极限刷怪".

    自然刷怪只改变出怪种类, 再由游戏内置的函数生成出怪列表.

    极限刷怪是把所选僵尸种类按顺序均匀地填充到出怪列表.

    @示例:

    >>> SetZombies(["撑杆", "舞王", "冰车", "海豚", "气球", "矿工", "跳跳", "扶梯", "白眼", "红眼"])
    """

    if mode in ("自然", "内置", "内置生成", "自然出怪", "自然刷怪"):
        set_internal_spawn(zombies + ["普僵"])
    elif mode in ("极限", "填充", "均匀填充", "极限出怪", "极限刷怪"):
        set_customize_spawn(zombies + ["旗帜"])
    else:
        error("未知刷怪模式: %s." % mode)


### 选卡

# (50, 160) 为左上角卡片中心坐标
# (215, 160) 为模仿者选卡界面左上角卡片中心坐标
# 单张卡片宽度约 50px 高度约 70px
# 模仿者卡片位置 (490, 550), 成功点击后等待界面出现再选卡
# 每次选完卡均等待一定时间

SEED_0_0_X = 50
SEED_0_0_Y = 160
IMITATER_SEED_0_0_X = 215
IMITATER_SEED_0_0_Y = 160
SEED_WIDTH = 50
SEED_HEIGHT = 70
IMITATER_X = 490
IMITATER_Y = 550
IMITATER_SHOW_UP = 0
SEED_DELAY_TIME = 5


def select_seed_by_crood(row, col, imitater=False):
    """
    选择单张卡片.

    @参数 row(int): 行

    @参数 col(int): 列

    @参数 imitater(bool): 是否为模仿者
    """

    if imitater:
        if row not in (1, 2, 3, 4, 5):
            critical("卡片行数 %d 超出有效范围." % row)
        if col not in (1, 2, 3, 4, 5, 6, 7, 8):
            critical("卡片列数 %d 超出有效范围." % col)
    else:
        if row not in (1, 2, 3, 4, 5, 6):
            critical("卡片行数 %d 超出有效范围." % row)
        if col not in (1, 2, 3, 4, 5, 6, 7, 8):
            critical("卡片列数 %d 超出有效范围." % col)

    if imitater:
        special_button_click(IMITATER_X, IMITATER_Y)
        thread_sleep_for(IMITATER_SHOW_UP)
        x = IMITATER_SEED_0_0_X + (col - 1) * (SEED_WIDTH + 1)
        y = IMITATER_SEED_0_0_Y + (row - 1) * (SEED_HEIGHT + 2)
    else:
        x = SEED_0_0_X + (col - 1) * (SEED_WIDTH + 3)
        y = SEED_0_0_Y + (row - 1) * (SEED_HEIGHT + 0)

    # 不小心点进了图鉴或者商店
    window_type = read_memory("int", 0x6A9EC0, 0x320, 0x88, 0xC)
    if window_type == 1:
        left_click(720, 580)
        thread_sleep_for(5)
    elif window_type == 4:
        left_click(430, 550)
        thread_sleep_for(5)

    special_button_click(x, y)
    thread_sleep_for(SEED_DELAY_TIME)

    if imitater:
        im_str = seeds_imitater_string[0] + " "
    else:
        im_str = ""
    seed_str = seeds_string[(row - 1) * 8 + (col - 1)][0]
    info("选择单张卡片 " + im_str + seed_str + ".")


@functools.singledispatch
def seed_to_crood(seed):
    """
    卡片转换为 (行, 列, 模仿者) 的标准形式.

    根据参数类型选择不同的实现.

    @参数 seed(int/tuple/str): 卡片

    @示例:

    >>> seed_to_crood(14 + 48)
    (2, 7, True)

    >>> seed_to_crood((2, 7, True))
    (2, 7, True)

    >>> seed_to_crood("复制冰")
    (2, 7, True)
    """
    error("卡片参数不支持 %s 类型." % type(seed))


@seed_to_crood.register(int)
def _(seed):
    if seed == 1437:
        row = 4
        col = 6
        imitater = False
    else:
        imitater = seed >= 48
        index = seed % 48
        row, col = divmod(index, 8)
    return row + 1, col + 1, imitater


@seed_to_crood.register(tuple)
def _(seed):
    if len(seed) == 2:
        row, col = seed
        imitater = False
    elif len(seed) == 3:
        row, col, im = seed
        imitater = im not in (False, 0)
    return row, col, imitater


@seed_to_crood.register(str)
def _(seed):
    if not seed in seeds_string_dict:
        error("未知卡片名称: %s." % seed)
    seed_index = seeds_string_dict[seed]  # 卡片代号(+48)
    imitater = seed_index >= 48
    index = seed_index % 48
    row, col = divmod(index, 8)
    return row + 1, col + 1, imitater


# 检查已选卡片是否完全相符
def slots_exact_match(seeds_selected):
    slots_count = read_memory("int", 0x6A9EC0, 0x768, 0x144, 0x24)
    slots_selected = read_memory("int", 0x6A9EC0, 0x774, 0xD24)
    if slots_selected < slots_count:
        return False
    match = True
    for i in range(slots_count):
        row, col, imitater = seeds_selected[i]
        seed_index = (row - 1) * 8 + (col - 1)
        if imitater:
            seed_index = 48
        seed_plant = read_memory("int", 0x6A9EC0, 0x774, 0xC4 + seed_index * 0x3C)
        seed_status = read_memory("int", 0x6A9EC0, 0x774, 0xC8 + seed_index * 0x3C)
        # !卡片对应的植物正确并且位于卡槽中
        if not (seed_plant == seed_index and seed_status == 1):
            match = False
            break
    return match


# 清空卡槽所有卡片
def clear_slots():
    slots_count = read_memory("int", 0x6A9EC0, 0x768, 0x144, 0x24)
    for slot_index in reversed(range(slots_count)):  # 逆序
        slot_index += 1
        if slots_count == 10:
            x = 63 + 51 * slot_index
        elif slots_count == 9:
            x = 63 + 52 * slot_index
        elif slots_count == 8:
            x = 61 + 54 * slot_index
        elif slots_count == 7:
            x = 61 + 59 * slot_index
        else:
            x = 61 + 59 * slot_index
        x -= 10
        y = 12
        special_button_click(x, y)
        thread_sleep_for(5)
    thread_sleep_for(5)


def select_all_seeds(seeds_selected=None):
    """
    选择所有卡片.
    """

    # 不小心点进了图鉴或者商店
    window_type = read_memory("int", 0x6A9EC0, 0x320, 0x88, 0xC)
    if window_type == 1:
        left_click(720, 580)
        thread_sleep_for(5)
    elif window_type == 4:
        left_click(430, 550)
        thread_sleep_for(5)

    default_seeds = [40, 41, 42, 43, 44, 45, 46, 47, 8, 8 + 48]
    slots_count = read_memory("int", 0x6A9EC0, 0x768, 0x144, 0x24)

    # 默认八张紫卡和两张免费卡
    if seeds_selected is None:
        seeds_selected = default_seeds[0:slots_count]

    # 参数个数小于卡槽数则用默认卡片填充
    while len(seeds_selected) < slots_count:
        for seed in default_seeds:
            if seed_to_crood(seed) not in [seed_to_crood(s) for s in seeds_selected]:
                seeds_selected += [seed]
                break

    if len(seeds_selected) > slots_count:
        critical("已选卡片数 %d 超过卡槽格数 %d." % (len(seeds_selected), slots_count))

    # 卡片列表转换为标准形式
    seeds_selected = [seed_to_crood(seed) for seed in seeds_selected]
    info("所选卡片转换为标准形式 %s." % seeds_selected)

    clear_slots()  # 清空卡槽已选卡片
    retry_count = 0

    while not slots_exact_match(seeds_selected):

        if retry_count > 3:
            critical("选卡重试多次失败.")
        retry_count += 1
        info("选卡过程未完成, 正在重试.")

        # 清空卡槽并选择所有卡片
        clear_slots()
        for seed in seeds_selected:
            row, col, imitater = seed
            select_seed_by_crood(row, col, imitater)
        thread_sleep_for(5)  # 等内存变化

    thread_sleep_for(5)  # 等内存变化


def lets_rock():
    while read_memory("bool", 0x6A9EC0, 0x768, 0x15C, 0x2C):  # 位于选卡界面
        left_down(234, 567)
        thread_sleep_for(1)
        left_up(234, 567)
        thread_sleep_for(10)
        while read_memory("int", 0x6A9EC0, 0x320, 0x94) != 0:  # 出现了对话框
            left_click(320, 400)
            thread_sleep_for(10)


def select_seeds_and_lets_rock(seeds_selected=None):
    """
    选卡并开始游戏.

    选择所有卡片, 点击开始游戏, 更新场景数据, 更新卡片列表, 更新加农炮列表, 等待开场红字消失.

    建议把鼠标光标移出窗口外以避免可能出现的模仿者选卡失败.

    @参数 seeds_selected(list): 卡片列表, 长度不大于卡槽格数.

    列表为空默认选择八张紫卡和两张免费卡, 卡片个数小于卡槽数则用默认卡片填充.

    单张卡片 seed 可用 int/tuple/str 表示, 不同的表示方法可混用.

    seed(int): 卡片序号, 0 为豌豆射手, 47 为玉米加农炮, 对于模仿者这个数字再加上 48.

    seed(tuple): 卡片位置, 用 (行, 列, 是否模仿者) 表示, 第三项可省略, 默认非模仿者.

    seed(str): 卡片名称, 参考 seeds_string, 包含了一些常用名字.

    @示例:

    >>> SelectCards()

    >>> SelectCards([14, 14 + 48, 17, 2, 3, 30, 33, 13, 9, 8])

    >>> SelectCards([(2, 7), (2, 7, True), (3, 2), (1, 3, False), (1, 4, False), (4, 7), (5, 2), (2, 6), (2, 2), (2, 1),])

    >>> SelectCards(["寒冰菇", "复制冰", "窝瓜", "樱桃", "坚果", "南瓜", "花盆", "胆小", "阳光", "小喷"])

    >>> SelectCards(["小喷菇", "模仿者小喷菇"])
    """
    gc.collect()

    # 激活并提高运行优先级
    set_pvz_foreground()
    set_pvz_high_priority()

    # 等待战斗结束进入选卡界面
    while game_ui() != 2:
        thread_sleep_for(1)

    # 选卡
    select_all_seeds(seeds_selected)
    lets_rock()

    # 更新相关数据
    update_game_scene()
    update_seeds_list()
    update_cob_cannon_list()

    # 选卡后等待直至正式开始战斗
    while game_ui() != 3:
        thread_sleep_for(1)


### 获取卡槽信息

# 每张卡片在卡槽里的位置, 用于根据卡片代号找卡槽位置
seeds_in_slot = [None] * (48 * 2)

# 卡槽中每张卡片的代号, 用于根据卡槽位置找卡片代号
slot_seeds = [None] * 10


def update_seeds_list():
    """
    更新卡片相关数据. 该函数须在点击 "Let's Rock!" 后调用.
    """
    global seeds_in_slot, slot_seeds
    seeds_in_slot = [None] * (48 * 2)
    slot_seeds = [None] * 10

    slots_count = read_memory("int", 0x6A9EC0, 0x768, 0x144, 0x24)
    slots_offset = read_memory("unsigned int", 0x6A9EC0, 0x768, 0x144)
    for i in range(slots_count):
        seed_type = read_memory("int", slots_offset + 0x5C + i * 0x50)
        seed_imitater_type = read_memory("int", slots_offset + 0x60 + i * 0x50)
        if seed_type == 48:
            seed = seed_imitater_type + 48
        else:
            seed = seed_type
        seeds_in_slot[seed] = i + 1
        slot_seeds[i] = seed

    # info("更新卡片位置 %s." % str(seeds_in_slot))
    info("更新卡槽代号 %s." % str(slot_seeds))


# 卡片名字 name
# 卡片代号 seed
# 卡槽位置 index


# (name: str) -> int
def get_seed_by_name(name):
    """
    根据卡片名字得到卡片代号. (模仿者 +48)
    """
    if name not in seeds_string_dict:
        error("未知卡片名称: %s." % name)
    return seeds_string_dict[name]


# (seed: int) -> int
def get_index_by_seed(seed):
    """
    根据卡片代号得到卡槽位置. 不在返回 None.
    """
    if seed not in range(48 * 2):
        error("卡片代号 %d 超出有效范围." % seed)
    return seeds_in_slot[seed]


# (name: str) -> int
def get_index_by_name(name):
    """
    根据卡片名字得到卡槽位置. 不在返回 None.
    """
    if name not in seeds_string_dict:
        error("未知卡片名称: %s." % name)
    return seeds_in_slot[seeds_string_dict[name]]


# (index: int) -> int
def get_seed_by_index(index):
    """
    根据卡槽位置得到卡片代号.
    """
    if index not in range(1, 11):
        error("卡槽位置 %d 超出有效范围." % index)
    return slot_seeds[index - 1]


### 场景相关信息

# 卡槽格数, 选卡和用卡函数需要
slots_count = 10

# 场景地图, 点击场上格子相关函数需要
game_scene = 2

scenes = {
    0: "Day",
    1: "Night",
    2: "Pool",
    3: "Fog",
    4: "Roof",
    5: "Moon",
    6: "Mushroom Garden",
    7: "Zen Garden",
    8: "Aquarium Garden",
    9: "Tree of Wisdom",
}


# 更新卡槽格数和场景地图
def update_game_scene():
    global slots_count, game_scene
    slots_count = read_memory("int", 0x6A9EC0, 0x768, 0x144, 0x24)
    game_scene = read_memory("int", 0x6A9EC0, 0x768, 0x554C)
    info("更新卡槽格数 %d." % slots_count)
    info("更新场景地图 %s." % scenes[game_scene])


### 场景点击操作

# 唯一内置鼠标锁
mouse_lock = threading.Lock()


def get_mouse_lock():
    """
    获取鼠标锁, 进行完整的 (不可分割的) 鼠标操作前加锁, 操作完毕后释放.

    @返回值 (object): 唯一内置鼠标锁.

    @示例:

    >>> MouseLock().acquire()  # 获取鼠标操作权
    >>> SafeClick()            # 安全右键避免冲突
    >>> pass                   # 干点什么
    >>> MouseLock().release()  # 释放鼠标操作权

    >>> with MouseLock():  # 获取鼠标操作权, 代码块结束后自动释放
    >>>     SafeClick()    # 安全右键避免冲突
    >>>     pass           # 干点什么
    """

    return mouse_lock


def safe_click():
    """
    安全右键.

    即右键单击左上角, 用于取消之前的 (可能未完成的) 操作以避免冲突.
    """
    right_click(0, 0)


def click_seed(seed):
    """
    点击卡槽中的卡片.

    @参数 seed(int/str): 卡槽第几格或者卡片名称.

    @示例:

    >>> ClickSeed(5)  # 点击第 5 格卡槽

    >>> ClickSeed("樱桃")  # 点击卡槽中的樱桃卡片
    """

    if isinstance(seed, str):
        slot_index = get_index_by_name(seed)
        if slot_index is None:
            error("卡槽当中没有 %s 卡片, 操作失败." % seed)
    else:  # int
        slot_index = seed
        if slot_index not in range(1, 11):
            error("卡槽格数 %d 超出有效范围, 操作失败." % slot_index)

    if slots_count == 10:
        x = 63 + 51 * slot_index
    elif slots_count == 9:
        x = 63 + 52 * slot_index
    elif slots_count == 8:
        x = 61 + 54 * slot_index
    elif slots_count == 7:
        x = 61 + 59 * slot_index
    else:
        x = 61 + 59 * slot_index
    y = 12
    left_click(x, y)


def click_shovel():
    """
    点击铲子.
    """
    if slots_count == 10:
        x = 640
    elif slots_count == 9:
        x = 600
    elif slots_count == 8:
        x = 570
    elif slots_count == 7:
        x = 550
    else:
        x = 490
    y = 36
    left_click(x, y)


# 坐标转换
def rc2xy(*crood):
    """
    row, col -> x, y
    """

    if isinstance(crood[0], tuple):
        row, col = crood[0]
    else:
        row, col = crood

    x = 80 * col
    if game_scene in (2, 3):
        y = 55 + 85 * row
    elif game_scene in (4, 5):
        if col >= 6:
            y = 45 + 85 * row
        else:
            y = 45 + 85 * row + 20 * (6 - col)
    else:
        y = 40 + 100 * row

    return int(x), int(y)  # 取整


def click_grid(*crood):
    """
    点击场上格点.

    @参数 crood(float/tuple): 坐标, 两个分别表示 行/列 的数字或者一个 (行, 列) 元组, 数字可为小数.

    @示例:

    >>> ClickGrid((2, 9))  # 点击 2 行 9 列

    >>> ClickGrid(2, 9)  # 同上
    """
    x, y = rc2xy(*crood)
    left_click(x, y)


### 更新炮列表

# 炮列表 (row, col) x n
cob_list = []

# 用炮序号
cob_index = 0

# 炮列表锁
cob_lock = threading.Lock()


def update_cob_cannon_list(cobs=None):
    """
    更新玉米加农炮列表.

    选卡时会自动调用, 空参数则自动找炮. 若需要自定义顺序请在选卡函数后面使用.

    @参数 cobs(list): 加农炮列表, 包括若干个 (行, 列) 元组, 以后轮坐标为准.

    @示例:

    >>> UpdatePaoList()

    >>> UpdatePaoList([(3, 1), (4, 1), (3, 3), (4, 3), (1, 5), (2, 5), (3, 5), (4, 5), (5, 5), (6, 5)])

    >>> UpdatePaoList(
    >>>     [
    >>>         (r, c)
    >>>         for r in (1, 2, 3, 4, 5, 6)
    >>>         for c in (1, 3, 5, 7)
    >>>         if not (r in (3, 4) and c == 7)
    >>>     ]
    >>> )

    >>> UpdatePaoList([
    >>>                     (1, 5), (1, 7),
    >>>     (2, 1),         (2, 5), (2, 7),
    >>>     (3, 1), (3, 3), (3, 5), (3, 7),
    >>>     (4, 1), (4, 3), (4, 5), (4, 7),
    >>>     (5, 1),         (5, 5), (5, 7),
    >>>                     (6, 5), (6, 7),
    >>>     ])
    """

    global cob_list, cob_index

    cob_list_tmp = []
    plants_count_max = read_memory("unsigned int", 0x6A9EC0, 0x768, 0xB0)
    plants_offset = read_memory("unsigned int", 0x6A9EC0, 0x768, 0xAC)
    for i in range(plants_count_max):
        plant_dead = read_memory("bool", plants_offset + 0x141 + 0x14C * i)
        plant_crushed = read_memory("bool", plants_offset + 0x142 + 0x14C * i)
        plant_type = read_memory("int", plants_offset + 0x24 + 0x14C * i)
        if not plant_dead and not plant_crushed and plant_type == 47:
            cob_row = read_memory("int", plants_offset + 0x1C + 0x14C * i)
            cob_col = read_memory("int", plants_offset + 0x28 + 0x14C * i)
            cob_list_tmp.append((cob_col + 1, cob_row + 1))  # 优先按列排
    cob_list_tmp.sort()  # 排序
    cob_list_tmp = [(r, c) for c, r in cob_list_tmp]  # 再反过来

    cob_lock.acquire()  # 加锁

    cob_list = []
    cob_index = 0

    # 自动查找
    if cobs is None:
        cob_list = cob_list_tmp
        info("查找场上玉米炮 %s." % str(cob_list))

    # 手动更新
    else:
        cobs_not_exist = []
        for cob in cobs:
            if cob not in cob_list_tmp:
                cobs_not_exist.append(cob)
        if len(cobs_not_exist) > 0:
            error("玉米炮 %s 不存在." % str(cobs_not_exist))
        cob_list = cobs
        info("手动更新炮列表 %s." % str(cob_list))

    cob_lock.release()  # 解锁


### 用卡操作


def use_seed(seed, *crood):
    """
    用卡操作.

    @参数 seed(int/str): 卡槽第几格或者卡片名称.

    @参数 crood(int/tuple): 坐标, 两个分别表示 行/列 的数字或者一个 (行, 列) 元组, 数字均为整数.

    @示例:

    >>> Card(1, (2, 3))  # 将卡槽中的第 1 张卡片种在 2 行 3 列

    >>> Card(1, 2, 3)  # 同上

    >>> Card("樱桃", (5, 9))  # 将樱桃种在 5 行 9 列

    >>> Card("樱桃", 5, 9)  # 同上
    """

    if isinstance(seed, str):
        seed_type = get_seed_by_name(seed)
        slot_index = get_index_by_seed(seed_type)
    else:  # int
        seed_type = get_seed_by_index(seed)
        slot_index = seed

    # 墓碑/咖啡豆 理想种植坐标偏上约 30px
    if seed_type in (11, 35, 11 + 48, 35 + 48):
        row_fix = -0.3
    else:
        row_fix = 0
    if isinstance(crood[0], tuple):
        row, col = crood[0]
    else:
        row, col = crood
    row += row_fix

    mouse_lock.acquire()
    safe_click()
    click_seed(slot_index)
    click_grid((row, col))
    safe_click()
    mouse_lock.release()

    if isinstance(seed, str):
        info("向 %s 种植 %s 卡片." % (str(crood), str(seed)))
    else:  # int
        info("向 %s 种植卡槽第 %s 张卡片." % (str(crood), str(seed)))


### 用铲子操作


def use_shovel(*croods):
    """
    用铲子操作.

    @参数 croods(float/tuple): 坐标, 两个分别表示 行/列 的数字或者一至多个 (行, 列) 元组, 数字可为小数.

    @示例:

    >>> Shovel((3, 4))  # 铲掉 3 行 4 列的普通植物

    >>> Shovel(3, 4)  # 同上

    >>> Shovel((5 + 0.1, 6))  # 铲掉 5 行 6 列的南瓜头

    >>> Shovel((1, 9), (2, 9), (5, 9), (6, 9))  # 铲掉所有 9 列垫材
    """

    mouse_lock.acquire()
    safe_click()

    if isinstance(croods[0], tuple):
        for crood in croods:
            click_shovel()
            click_grid(crood)
    else:  # float/int
        click_shovel()
        click_grid(*croods)

    safe_click()
    mouse_lock.release()

    info("对格子 %s 使用铲子." % str(croods))


### 用炮操作


@functools.singledispatch
def fire_cob(*croods):
    """
    用炮操作.

    @参数 croods(float/tuple/list): 落点, 一至多个格式为 (行, 列) 的元组, 或者一个包含了这些元组的列表.

    @示例:

    >>> Pao((2, 9))

    >>> Pao(2, 9)

    >>> Pao((2, 9), (5, 9))

    >>> Pao((2, 9), (5, 9), (2, 9), (5, 9))

    >>> Pao([(2, 9), (5, 9), (2, 9), (5, 9)])
    """
    error("炮落点参数格式不正确.")


@fire_cob.register(int)
def _(fall_row, fall_col):
    global cob_list, cob_index
    cob_count = len(cob_list)
    if cob_count == 0:
        error("炮列表为空.")

    cob_lock.acquire()
    cob_row, cob_col = cob_list[cob_index]
    fire_cob_by_crood(cob_row, cob_col, fall_row, fall_col)
    cob_index += 1
    cob_index %= cob_count
    cob_lock.release()


@fire_cob.register(tuple)
def _(*fall_grids):
    global cob_list, cob_index
    cob_count = len(cob_list)
    if cob_count == 0:
        error("炮列表为空.")

    cob_lock.acquire()
    for grid in fall_grids:
        cob_row, cob_col = cob_list[cob_index]
        fall_row, fall_col = grid
        fire_cob_by_crood(cob_row, cob_col, fall_row, fall_col)
        cob_index += 1
        cob_index %= cob_count
    cob_lock.release()


@fire_cob.register(list)
def _(fall_grids):
    fire_cob(*fall_grids)


### 屋顶用炮操作

# 屋顶玉米炮飞行时间, 只考虑落点前场 7~9 列的情况
flying_time = {1: 359, 2: 362, 3: 364, 4: 367, 5: 369, 6: 372, 7: 373}


def get_cob_flying_time(cob_col, fall_col):
    # 暂不考虑 fall_col
    if cob_col in flying_time:
        return flying_time[cob_col]
    else:
        return 373


FLYING_TIME = 373


@running_in_thread
@functools.singledispatch
def fire_cob_on_roof(*croods):
    """
    屋顶修正飞行时间发炮. 参数格式与 `Pao()` 相同.

    此函数开新线程开销较大不适合精确键控, 只适用于前场落点 (约 7~9 列).
    """
    error("参数格式不正确.")


@fire_cob_on_roof.register(int)
def _(fall_row, fall_col):
    clock = game_clock()  # 参照时钟
    global cob_list, cob_index
    cob_count = len(cob_list)
    if cob_count == 0:
        error("炮列表为空.")

    cob_lock.acquire()
    cob_row, cob_col = cob_list[cob_index]
    cob_index += 1
    cob_index %= cob_count
    cob_lock.release()

    flying_time = get_cob_flying_time(cob_col, fall_col)
    while (game_clock() - clock) < (FLYING_TIME - flying_time):
        delay_a_little_time()
    fire_cob_by_crood(cob_row, cob_col, fall_row, fall_col)


@fire_cob_on_roof.register(tuple)
def _(*fall_grids):
    clock = game_clock()  # 参照时钟
    global cob_list, cob_index
    cob_count = len(cob_list)
    if cob_count == 0:
        error("炮列表为空.")

    # (flying_time, cob_row, cob_col, fall_row, fall_col) x n
    operate_list = []

    cob_lock.acquire()
    for grid in fall_grids:
        cob_row, cob_col = cob_list[cob_index]
        fall_row, fall_col = grid
        flying_time = get_cob_flying_time(cob_col, fall_col)
        operate_list.append((flying_time, cob_row, cob_col, fall_row, fall_col))
        cob_index += 1
        cob_index %= cob_count
    cob_lock.release()

    operate_list.sort()  # 根据飞行时间排序
    for op in reversed(operate_list):  # 逆序发射
        flying_time, cob_row, cob_col, fall_row, fall_col = op
        while (game_clock() - clock) < (FLYING_TIME - flying_time):
            delay_a_little_time()
        fire_cob_by_crood(cob_row, cob_col, fall_row, fall_col)


@fire_cob_on_roof.register(list)
def _(fall_grids):
    return fire_cob_on_roof(*fall_grids)


### 跳炮


def skip_cob_index(num):
    """
    按炮列表顺序跳过即将发射的一定数量的玉米炮, 通常用于 wave9/19 手动收尾.

    @参数 num(int): 数量.
    """
    global cob_list, cob_index

    cob_lock.acquire()
    cob_index += num
    cob_index %= len(cob_list)
    cob_lock.release()

    info("跳过炮列表中的 %d 门炮." % num)


### 直接发炮

# 炮身点击次数
CLICK_COUNT = 3


# 无视内置炮列表直接指定炮位和落点
def fire_cob_by_crood(cob_row, cob_col, fall_row, fall_col):

    mouse_lock.acquire()
    safe_click()
    for _ in range(CLICK_COUNT):
        # 点炮位置稍微偏离, 配合改内存解决炮粘手的问题
        click_grid(cob_row + 2 / 85, cob_col - 2 / 80)
    click_grid(fall_row, fall_col)
    safe_click()
    mouse_lock.release()

    info("从 (%d, %d) 向 (%d, %d) 发射玉米炮." % (cob_row, cob_col, fall_row, fall_col))


### 阻塞延时

# 当前波次刷新时间点
refresh_time_point = 0


def game_delay_for(time_cs):
    """
    游戏内部时钟延时. 相对于线程睡眠更准确.

    只能在战斗界面 `[[0x6A9EC0]+0x7FC] == 3` 使用, 游戏暂停时计时同样暂停.

    @参数 time_cs(int): 时间, 单位 cs, 精度 1.
    """

    if time_cs > 0:
        clock = game_clock()
        while (game_clock() - clock) < time_cs:
            delay_a_little_time()
    elif time_cs == 0:
        pass
    else:
        error("游戏延时参数不能小于零.")


def until_countdown(time_cs, hugewave=False):
    """
    等待直至刷新倒计时数值达到指定值.

    调用时需要保证上一波已经刷出. 该函数仅保留兼容旧式写法, 已不建议使用.

    @参数 time_cs(int): 倒计时数值, 单位 cs, 精度 1. 建议范围 [200, 1].

    第一波最早 599, 旗帜波最早 750.

    @参数 hugewave(bool): 是否为旗帜波, 默认不是. 可用 (波数 % 10 == 0) 判断.

    @示例:

    >>> Countdown(100)  # 非旗帜波 100cs 预判

    >>> Countdown(55, True)  # 旗帜波 55cs 预判

    >>> Countdown(95, wave % 10 == 0)  # 第 wave 波 95cs 预判
    """

    if not hugewave:
        while wave_countdown() > time_cs:
            delay_a_little_time()
    else:
        while wave_countdown() > 5:  # 这里用 5 而不用 4, 为了支持大波 750cs 预判
            delay_a_little_time()
        while huge_wave_countdown() > time_cs:
            delay_a_little_time()


# 倒计时(大波倒计时/初始倒计时)小于等于 200(750/599) 才算激活刷新
refresh_trigger = [750 if wave % 10 == 0 else 599 if wave == 1 else 200 for wave in range(1, 21)]


def until_relative_time_after_refresh(time_relative_cs, wave):
    """
    读内存获取刷新状况, 等待直至与设定波次刷新时间点的差值达到指定值.

    该函数须在每波操作开始时执行一次. 通常用于预判 (在设定波次刷新前调用), 也可以在设定波次刷新之后调用.

    @参数 time_relative_cs(int): 与刷新时间点的相对时间, 单位 cs, 精度 1. 建议范围 [-200, 400].

    第一波最早 -599, 旗帜波最早 -750. 为了方便可统一给每波设置类似 -180 等数值.

    因为脚本语言的精度问题, 设置成 -599/-750/-200 等过早的边界值时可能会因为实际达到时间已经超过该值而引起报错.

    @参数 wave(int): 波数. 用于内部判断刷新状况以及本波是否为旗帜波.

    @示例:

    >>> Prejudge(-100, wave)  # 第 wave 波刷新前 100cs 预判

    >>> Prejudge(-150, 20)  # 第 20 波预判炮炸珊瑚时机

    >>> Prejudge(300, wave)  # 第 wave 波刷新后 300cs

    >>> Prejudge(900 - 200 - 373, wave)  # 第 wave 波 900cs 波长激活炸时机
    """

    # 设定波次的刷新时间点
    global refresh_time_point

    _current_wave = current_wave()

    # 设定波次还未刷出
    if _current_wave < wave:

        # 等待设定预判波次的上一波刷出
        if _current_wave < wave - 1:
            while current_wave() < (wave - 1):
                delay_a_little_time()

        # 等到本波触发刷新
        is_huge_wave = wave % 10 == 0
        until_countdown(refresh_trigger[wave - 1], is_huge_wave)

        # 计算实际倒计时数值
        _wave_countdown = wave_countdown()
        _huge_wave_countdown = huge_wave_countdown()
        if is_huge_wave:
            if _wave_countdown in (4, 5):
                countdown = _huge_wave_countdown
            else:
                countdown = _wave_countdown - 5 + 750
        else:
            countdown = _wave_countdown

        # 计算刷新时间点(倒计时变为下一波初始值时)的时钟数值
        _game_clock = game_clock()
        refresh_time_point = _game_clock + countdown

        # 等待 目标相对时间 和 当前相对时间(即倒计时数值负值) 的差值
        time_to_wait = time_relative_cs + countdown
        if time_to_wait >= 0:
            game_delay_for(time_to_wait)
        else:
            error("第 %d 波设定时间 %d 已经过去, 当前相对时间 %d." % (wave, time_relative_cs, time_relative_cs - time_to_wait))

    # 设定波次已经刷出
    elif _current_wave == wave:

        # 获取当前时钟/倒计时数值/倒计时初始数值
        _game_clock = game_clock()
        _wave_countdown = wave_countdown()
        _wave_init_countdown = wave_init_countdown()

        if _wave_countdown <= 200:
            warning("设定波次 %d 的下一波即将刷新, 请调整脚本写法." % wave)

        # 计算刷新时间点(倒计时变为下一波初始值时)的时钟数值
        refresh_time_point = _game_clock - (_wave_init_countdown - _wave_countdown)

        # 等到设定时间
        time_to_wait = time_relative_cs - (_wave_init_countdown - _wave_countdown)
        if time_to_wait >= 0:
            game_delay_for(time_to_wait)
        else:
            error("第 %d 波设定时间 %d 已经过去, 当前相对时间 %d." % (wave, time_relative_cs, time_relative_cs - time_to_wait))

    # 设定波次的下一波已经刷出
    else:
        error("设定波次 %d 的下一波已经刷新, 请调整脚本写法." % wave)


def until_relative_time(time_relative_cs):
    """
    等待直至当前时间戳与本波刷新时间点的差值达到指定值.

    该函数需要配合 Prejudge() 使用. 多个 Until() 连用时注意调用顺序必须符合时间先后顺序.

    @参数 time_relative_cs(int): 相对时间, 单位 cs, 精度 1. 建议范围 [-200, 2300].

    @示例:

    >>> Until(-95)  # 刷新前 95cs

    >>> Until(180)  # 刷新后 180cs

    >>> Until(-150)  # 炮炸珊瑚可用时机

    >>> Until(444 - 373)  # 444cs 生效炮发射时机

    >>> Until(601 + 20 - 298)  # 加速波下一波 20cs 预判冰点咖啡豆时机

    >>> Until(601 + 5 - 100 - 320)  # 加速波下一波 5cs 预判冰复制冰种植时机

    >>> Until(1200 - 200 - 373)  # 1200cs 波长激活炸时机

    >>> Until(4500 - 5)  # 收尾波拖满时红字出现时机

    >>> Until(5500)  # 最后一大波白字出现时机
    """

    # while (game_clock() - refresh_time_point) < time_relative_cs:
    #     delay_a_little_time()

    time_to_wait = time_relative_cs - (game_clock() - refresh_time_point)
    if time_to_wait >= 0:
        game_delay_for(time_to_wait)
    else:
        error("设定时间 %d 已经过去, 当前相对时间 %d." % (time_relative_cs, time_relative_cs - time_to_wait))


### 自动收集线程

collect_items_dict = {
    "银币": 1,
    "金币": 2,
    "钻石": 3,
    "阳光": 4,
    "小阳光": 5,
    "大阳光": 6,
    "太阳": 4,
    "小太阳": 5,
    "大太阳": 6,
    "幼苗": 17,
    "花苗": 17,
    "盆栽": 17,
    "花盆": 17,
    "礼盒": 17,
    "礼品盒": 17,
    1: 1,
    2: 2,
    3: 3,
    4: 4,
    5: 5,
    6: 6,
    17: 17,
}

item_type_names = {1: "银币", 2: "金币", 3: "钻石", 4: "阳光", 5: "小阳光", 6: "大阳光", 17: "幼苗"}


@running_in_thread
def auto_collect(collect_items=None, interval_cs=12):
    """
    自动收集场上资源, 在单独的子线程运行.

    为了避免操作冲突, 当鼠标选中 卡片/铲子/玉米炮 时会暂停收集.

    建议把鼠标光标移出窗口外以避免可能出现的游戏卡顿.

    @参数 collect_items(list[str/int]): 包含需要收集的资源类型的列表, 默认所有.

    可选值物品名称 ["银币", "金币", "钻石", "阳光", "小阳光", "大阳光", "幼苗"] 或者代号 [1, 2, 3, 4, 5, 6, 17].

    @参数 interval_cs(float/int): 点击间隔, 单位 cs, 默认 12.

    @示例:

    >>> AutoCollect()  # 自动收集所有资源

    >>> AutoCollect(["钻石", "阳光", "小阳光", "大阳光"], 20)  # 只收集钻石和各种阳光, 间隔 0.2s
    """

    if collect_items is None:
        collect_items = ["银币", "金币", "钻石", "阳光", "小阳光", "大阳光", "幼苗"]
    collect_items_list = [collect_items_dict[item] for item in collect_items]

    while game_ui() != 3:
        thread_sleep_for(10)

    info("启动自动收集线程.")

    while game_ui() == 3:
        items_count = read_memory("int", 0x6A9EC0, 0x768, 0xF4)
        items_count_max = read_memory("int", 0x6A9EC0, 0x768, 0xE8)
        items_offset = read_memory("int", 0x6A9EC0, 0x768, 0xE4)

        if items_count == 0:
            thread_sleep_for(interval_cs)
            continue

        uncollected_item_count = 0
        for i in range(items_count_max):
            disappeared = read_memory("bool", items_offset + 0x38 + 0xD8 * i)
            collected = read_memory("bool", items_offset + 0x50 + 0xD8 * i)
            item_type = read_memory("int", items_offset + 0x58 + 0xD8 * i)
            if not disappeared and not collected and item_type in collect_items_list:
                uncollected_item_count += 1
        if uncollected_item_count == 0:
            thread_sleep_for(interval_cs * 10)  # 等久一点
            continue

        for i in range(items_count_max):
            if game_ui() != 3:
                break

            # while game_paused():  # 一直收集
            # while game_paused() or mouse_in_game():  # 鼠标移出时收集
            while game_paused() or (mouse_in_game() and mouse_have_something()):  # 没选中卡炮铲时收集
                thread_sleep_for(interval_cs)

            disappeared = read_memory("bool", items_offset + 0x38 + 0xD8 * i)
            collected = read_memory("bool", items_offset + 0x50 + 0xD8 * i)
            item_type = read_memory("int", items_offset + 0x58 + 0xD8 * i)
            if not disappeared and not collected and item_type in collect_items_list:

                item_x = read_memory("float", items_offset + 0x24 + 0xD8 * i)
                item_y = read_memory("float", items_offset + 0x28 + 0xD8 * i)
                if item_x >= 0.0 and item_y >= 70.0:
                    # write_memory("bool", True, items_offset + 0x50 + 0xd8 * i)
                    x, y = int(item_x + 30), int(item_y + 30)
                    mouse_lock.acquire()
                    safe_click()
                    left_click(x, y)
                    safe_click()
                    mouse_lock.release()

                    debug("收集位于 (%d, %d) 的物品 %s." % (x, y, item_type_names[item_type]))
                    thread_sleep_for(interval_cs)
                    # thread_sleep_for(random.randint(int(interval_cs * 0.5), int(interval_cs * 1.5)))  # 时间波动

    info("停止自动收集线程.")


### 自动存冰线程


def get_seeds_index(seed):
    """
    获取某种卡片在卡槽里的所有位置.

    @参数 seed(str): 卡片名称.

    @返回值 (list[int]): 某种卡片 (包括其模仿者) 在卡槽的数组下标列表.
    """
    seed = get_seed_by_name(seed)
    seed %= 48

    seed_indexes = []
    slots_count = read_memory("int", 0x6A9EC0, 0x768, 0x144, 0x24)
    slots_offset = read_memory("unsigned int", 0x6A9EC0, 0x768, 0x144)
    for i in range(slots_count):
        seed_type = read_memory("int", slots_offset + 0x5C + i * 0x50)
        seed_imitater_type = read_memory("int", slots_offset + 0x60 + i * 0x50)
        if seed_type == seed or (seed_type == 48 and seed_imitater_type == seed):
            seed_indexes.append(i)
    return seed_indexes


def get_plants_croods():
    """
    获取场上植物坐标.
    """
    croods = []
    plants_count_max = read_memory("unsigned int", 0x6A9EC0, 0x768, 0xB0)
    plants_offset = read_memory("unsigned int", 0x6A9EC0, 0x768, 0xAC)
    for i in range(plants_count_max):
        plant_dead = read_memory("bool", plants_offset + 0x141 + 0x14C * i)
        plant_crushed = read_memory("bool", plants_offset + 0x142 + 0x14C * i)
        if not plant_dead and not plant_crushed:
            plant_type = read_memory("int", plants_offset + 0x24 + 0x14C * i)
            plant_row = read_memory("int", plants_offset + 0x1C + 0x14C * i)
            plant_col = read_memory("int", plants_offset + 0x28 + 0x14C * i)
            croods.append((plant_type, plant_row + 1, plant_col + 1))
    return croods


def get_block_type(*crood):
    """
    获取格子类型. 1.lawn 2.bare 3.pool
    """
    if isinstance(crood[0], tuple):
        row, col = crood[0]
    else:
        row, col = crood
    row, col = row - 1, col - 1
    return read_memory("int", 0x6a9ec0, 0x768, 0x168 + row * 0x04 + col * 0x18)


# 存冰位
ice_spots = []
ice_total = 0


@running_in_thread
def auto_fill_ice(spots=None, total=0x7FFFFFFF):
    """
    设置存冰位置和要存的数量, 将会在单独的子线程运行自动存冰.

    @参数 spots(list): 存冰点, 包括若干个 (行, 列) 元组. 临时位在后. 默认为场上现有存冰的位置.

    @参数 total(int): 总个数, 默认无限.

    @示例:

    >>> IceSpots()

    >>> IceSpots([(6, 1), (5, 1), (2, 1), (1, 1)], 10)  # 往指定位置总计存 10 个冰
    """

    while game_ui() != 3:
        thread_sleep_for(1)

    info("启动自动存冰线程.")

    # 默认为场上现有存冰的位置
    if spots is None:
        spots = []
        plants = get_plants_croods()
        for plant_type, plant_row, plant_col in plants:
            if plant_type == get_seed_by_name("寒冰菇"):
                spots.append((plant_row, plant_col))
        if spots == []:
            error("场上没有寒冰菇, 退出自动存冰.")
            return

    global ice_spots, ice_total
    ice_spots = spots
    ice_total = total

    slots_offset = read_memory("unsigned int", 0x6A9EC0, 0x768, 0x144)

    ice_seeds_index = get_seeds_index("寒冰菇")  # 获取所有寒冰菇卡片的下标
    if ice_seeds_index == []:
        error("卡槽没有寒冰菇, 退出自动存冰.")

    filled = 0  # 已存数量
    while game_ui() == 3 and filled < total:

        while game_paused():
            thread_sleep_for(1)  # 等待暂停取消
        if game_ui() != 3:
            break

        croods_which_has_plant = []
        plants = get_plants_croods()
        for plant_type, plant_row, plant_col in plants:
            if plant_type not in (16, 30, 33):  # 睡莲/南瓜/花盆
                croods_which_has_plant.append((plant_row, plant_col))
                if plant_type == 47:  # 玉米炮占两格
                    croods_which_has_plant.append((plant_row, plant_col + 1))
        ice_spot_which_has_plant = [i for i in ice_spots if i in croods_which_has_plant]

        # 这句忘了有什么作用了 ...
        if game_ui() != 3 and ice_spot_which_has_plant == []:
            break

        if set(ice_spot_which_has_plant) >= set(spots):  # 存冰位植物满了
            ice_seeds_cd_left = []
            for i in ice_seeds_index:
                seed_usable = read_memory("bool", slots_offset + 0x70 + i * 0x50)
                seed_cd_past = read_memory("int", slots_offset + 0x4C + i * 0x50)
                seed_cd_total = read_memory("int", slots_offset + 0x50 + i * 0x50)
                ice_seeds_cd_left.append(0 if seed_usable else (seed_cd_total - seed_cd_past))
            if min(ice_seeds_cd_left) > 0:  # 冰卡都在冷却时等待最小的卡片 CD
                info("寒冰菇卡片冷却中, 等待 %d." % (min(ice_seeds_cd_left) + 1))
                game_delay_for(min(ice_seeds_cd_left) + 1)
                continue
            else:  # TODO 冰卡可用时等待正在用的咖啡豆
                thread_sleep_for(2)  # 延时以减小遍历植物的 CPU 消耗
                continue

        # 遍历指定的存冰位
        for spot in spots:
            if game_ui() != 3:
                break

            # 如果该位置无植物则尝试存冰
            if pvz_ver() == "1.0.0.1051":
                seed_ice_cost = read_memory("int", 0x69F2C0 + 14 * 0x24)
            else:
                seed_ice_cost = read_memory("int", 0x69F2D0 + 14 * 0x24)
            sun = read_memory("int", 0x6A9EC0, 0x768, 0x5560)
            if sun < seed_ice_cost:
                thread_sleep_for(1)
                continue
            block_type = get_block_type(spot)
            # 1.草地 2.裸地 3.泳池 16.睡莲 33.花盆
            if (spot not in ice_spot_which_has_plant \
                and sun >= seed_ice_cost \
                and ((block_type == 1 and game_scene not in (4, 5)) \
                    or (block_type == 1 and game_scene in (4, 5) and (33, spot[0], spot[1]) in plants) \
                    or (block_type == 3 and (16, spot[0], spot[1]) in plants))):
                # 遍历寒冰菇卡片, 通常为 原版冰 x 1 + 复制冰 x 1
                for i in ice_seeds_index:
                    seed_usable = read_memory("bool", slots_offset + 0x70 + i * 0x50)
                    if seed_usable:
                        while game_paused():
                            thread_sleep_for(1)  # 等待暂停取消
                        mouse_lock.acquire()
                        safe_click()
                        click_seed(i + 1)
                        click_grid(spot)
                        safe_click()
                        mouse_lock.release()
                        filled += 1
                        info("往 %s 存冰 (第 %d 张卡)." % (str(spot), i + 1))
                        game_delay_for(1)  # 等待内存数据更新
                        break
                    else:
                        thread_sleep_for(1)
                break  # 不管有没有成功都重新遍历存冰位以保证顺序(先永久位后临时位)
            else:
                pass

    info("停止自动存冰线程.")


def activate_ice():
    """
    点冰. 使用咖啡豆激活存冰, 优先点临时位.

    该函数需要配合自动存冰线程 IceSpots() 使用.
    """
    coffee_index = get_index_by_seed(35)  # 咖啡豆位置
    if coffee_index is None:
        error("卡槽没有咖啡豆, 点冰失败.")

    mouse_lock.acquire()
    safe_click()
    click_seed(coffee_index)
    for spot in reversed(ice_spots):  # 优先点临时位
        row, col = spot
        row -= 0.3  # 咖啡豆 理想种植坐标偏上约 30px
        click_grid((row, col))
    safe_click()
    mouse_lock.release()
